"""Utils for plotting whole network graph."""

import collections
import itertools
from numbers import Number
from typing import Dict, List, Optional, Tuple, Union

import matplotlib
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
from toolz import valfilter

from . import plt_utils


class WholeNetworkFigure:
    """
    Class for creating a whole network figure.

    Attributes:
        nodes (pd.DataFrame): DataFrame containing node information.
        edges (pd.DataFrame): DataFrame containing edge information.
        layout (Dict[str, str]): Dictionary mapping node types to layout positions.
        cell_types (List[str]): List of unique cell types.
        video (bool): Whether to include video node.
        rendering (bool): Whether to include rendering node.
        motion_decoder (bool): Whether to include motion decoder node.
        decoded_motion (bool): Whether to include decoded motion node.
        pixel_accurate_motion (bool): Whether to include pixel-accurate motion node.
    """

    def __init__(
        self,
        connectome,
        video: bool = False,
        rendering: bool = False,
        motion_decoder: bool = False,
        decoded_motion: bool = False,
        pixel_accurate_motion: bool = False,
    ):
        self.nodes = connectome.nodes.to_df()
        self.edges = connectome.edges.to_df()
        self.layout = dict(connectome.layout[:].astype(str))
        self.cell_types = connectome.unique_cell_types[:].astype(str)

        layout = {}
        if video:
            layout.update({"video": "cartesian"})
        if rendering:
            layout.update({"rendering": "hexagonal"})
        layout.update(dict(connectome.layout[:].astype(str)))
        if motion_decoder:
            layout.update({"motion decoder": "decoder"})
        if decoded_motion:
            layout.update({"decoded motion": "motion"})
        if pixel_accurate_motion:
            layout.update({"pixel-accurate motion": "motion"})
        self.layout = layout
        self.video = video
        self.rendering = rendering
        self.motion_decoder = motion_decoder
        self.decoded_motion = decoded_motion
        self.pixel_accurate_motion = pixel_accurate_motion

    def init_figure(
        self,
        figsize: List[int] = [15, 6],
        fontsize: int = 6,
        decoder_box: bool = True,
        cell_type_labels: bool = True,
        neuropil_labels: bool = True,
        network_layout_axes_kwargs: Dict = {},
        add_graph_kwargs: Dict = {},
    ) -> None:
        """
        Initialize the figure with various components.

        Args:
            figsize: Size of the figure.
            fontsize: Font size for labels.
            decoder_box: Whether to add a decoder box.
            cell_type_labels: Whether to add cell type labels.
            neuropil_labels: Whether to add neuropil labels.
            network_layout_axes_kwargs: Additional kwargs for network_layout_axes.
            add_graph_kwargs: Additional kwargs for add_graph.
        """
        self.fig, self.axes, self.axes_centers = network_layout_axes(
            self.layout, figsize=figsize, **network_layout_axes_kwargs
        )
        self.ax_dict = {ax.get_label(): ax for ax in self.axes}
        self.add_graph(**add_graph_kwargs)

        self.add_retina_box()

        if decoder_box:
            self.add_decoded_box()

        if cell_type_labels:
            self.add_cell_type_labels(fontsize=fontsize)

        if neuropil_labels:
            self.add_neuropil_labels(fontsize=fontsize)

        if self.motion_decoder:
            self.add_decoder_sketch()

        self.add_arrows()

    def add_graph(
        self,
        edge_color_key: Optional[str] = None,
        arrows: bool = True,
        edge_alpha: float = 1.0,
        edge_width: float = 1.0,
        constant_edge_width: Optional[float] = 0.25,
        constant_edge_color: str = "#c5c5c5",
        edge_cmap: Optional[str] = None,
        nx_kwargs: Dict = {},
    ) -> None:
        """
        Add the graph to the figure.

        Args:
            edge_color_key: Key for edge color.
            arrows: Whether to add arrows to edges.
            edge_alpha: Alpha value for edges.
            edge_width: Width of edges.
            constant_edge_width: Constant width for all edges.
            constant_edge_color: Constant color for all edges.
            edge_cmap: Colormap for edges.
            nx_kwargs: Additional kwargs for networkx drawing.
        """

        def _network_graph(nodes, edges):
            """Transform graph from df to list to create networkx.Graph object."""
            nodes = nodes.groupby(by=["type"], sort=False, as_index=False).first().type
            edges = list(
                map(
                    lambda x: x.split(","),
                    (edges.source_type + "," + edges.target_type).unique(),
                )
            )
            return nodes, edges

        axes = {
            cell_type: [ax for ax in self.axes if ax.get_label() == cell_type][0]
            for cell_type in self.cell_types
        }

        (
            (lefts, bottoms, rights, tops),
            (
                centers,
                widths,
                height,
            ),
        ) = plt_utils.get_ax_positions(list(axes.values()))
        edge_ax = self.fig.add_axes([
            lefts.min(),
            bottoms.min(),
            rights.max() - lefts.min(),
            tops.max() - bottoms.min(),
        ])
        edge_ax.set_zorder(0)
        edge_ax = plt_utils.rm_spines(edge_ax, rm_xticks=True, rm_yticks=True)
        edge_ax.patch.set_alpha(0.0)
        edge_ax.set_ylim(0, 1)
        edge_ax.set_xlim(0, 1)

        fig_to_edge_ax = self.fig.transFigure + edge_ax.transData.inverted()
        positions = {
            key: fig_to_edge_ax.transform(value)
            for key, value in self.axes_centers.items()
        }

        nodes, edge_list = _network_graph(self.nodes, self.edges)

        if edge_color_key is not None and not constant_edge_color:
            grouped = self.edges.groupby(
                by=["source_type", "target_type"], sort=False, as_index=False
            ).mean(numeric_only=True)
            edge_color = {
                (row.source_type, row.target_type): row.sign
                for i, row in grouped.iterrows()
            }
            _edge_color = np.array(list(edge_color.values()))
            edge_vmin = -np.max(_edge_color) if np.any(_edge_color < 0) else 0
            edge_vmax = np.max(_edge_color)
        else:
            edge_color = {tuple(edge): constant_edge_color for edge in edge_list}
            edge_vmin = None
            edge_vmax = None

        grouped = self.edges.groupby(
            by=["source_type", "target_type"], sort=False, as_index=False
        ).mean(numeric_only=True)

        if constant_edge_width is None:
            edge_width = {
                (row.source_type, row.target_type): edge_width * (np.log(row.n_syn) + 1)
                for i, row in grouped.iterrows()
            }
        else:
            edge_width = {
                (row.source_type, row.target_type): constant_edge_width
                for i, row in grouped.iterrows()
            }

        graph = nx.DiGraph()
        graph.add_nodes_from(nodes)
        graph.add_edges_from(edge_list)

        draw_networkx_edges(
            graph,
            pos=positions,
            ax=edge_ax,
            edge_color=np.array([edge_color[tuple(edge)] for edge in edge_list]),
            edge_cmap=edge_cmap,
            edge_vmin=edge_vmin,
            edge_vmax=edge_vmax,
            alpha=edge_alpha,
            arrows=arrows,
            arrowstyle=(
                "-|>, head_length=0.4, head_width=0.075, widthA=1.0, "
                "widthB=1.0, lengthA=0.2, lengthB=0.2"
            ),
            width=np.array([edge_width[tuple(edge)] for edge in edge_list]),
            **nx_kwargs,
        )
        self.edge_ax = edge_ax

    def add_retina_box(self):
        retina_node_types = valfilter(lambda v: v == "retina", self.layout)
        axes = {
            node_type: [ax for ax in self.axes if ax.get_label() == node_type][0]
            for node_type in retina_node_types
        }
        (
            (lefts, bottoms, rights, tops),
            (
                centers,
                widths,
                height,
            ),
        ) = plt_utils.get_ax_positions(list(axes.values()))
        retina_box_ax = self.fig.add_axes(
            [
                lefts.min(),
                bottoms.min(),
                rights.max() - lefts.min(),
                tops.max() - bottoms.min(),
            ],
            label="retina_box",
        )
        retina_box_ax.patch.set_alpha(0)
        plt_utils.rm_spines(retina_box_ax)
        self.ax_dict["retina box"] = retina_box_ax

    def add_decoded_box(self):
        output_cell_types = valfilter(lambda v: v == "output", self.layout)
        axes = {
            cell_type: [ax for ax in self.axes if ax.get_label() == cell_type][0]
            for cell_type in output_cell_types
        }
        (lefts, bottoms, rights, tops), _ = plt_utils.get_ax_positions(
            list(axes.values())
        )
        bottom, top = plt_utils.get_lims((bottoms, tops), 0.02)
        left, right = plt_utils.get_lims((lefts, rights), 0.01)
        decoded_box_ax = self.fig.add_axes(
            [
                left,
                bottom,
                right - left,
                top - bottom,
            ],
            label="decoded_box",
        )
        decoded_box_ax.patch.set_alpha(0)
        decoded_box_ax.spines["top"].set_visible(True)
        decoded_box_ax.spines["right"].set_visible(True)
        decoded_box_ax.spines["left"].set_visible(True)
        decoded_box_ax.spines["bottom"].set_visible(True)
        decoded_box_ax.set_xticks([])
        decoded_box_ax.set_yticks([])
        self.ax_dict["decoded box"] = decoded_box_ax

    def add_decoder_sketch(self):
        ax = self.ax_dict["motion decoder"]
        nodes = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]
        edges = [
            (1, 11),
            (2, 10),
            (2, 12),
            (3, 10),
            (3, 11),
            (3, 12),
            (4, 11),
            (4, 12),
            (4, 14),
            (5, 10),
            (5, 12),
            (5, 13),
            (6, 11),
            (6, 13),
            (6, 14),
            (7, 12),
            (7, 14),
            (8, 13),
            (9, 14),
            (10, 15),
            (11, 16),
            (12, 15),
            (13, 15),
            (13, 16),
            (14, 16),
        ]
        graph = nx.Graph()
        graph.add_nodes_from(nodes)
        graph.add_edges_from(edges)
        x = [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2]
        y = [9, 8, 7, 6, 5, 4, 3, 2, 1, 7.3, 6.3, 5.3, 4.3, 3.3, 5.7, 4.7]
        x, y, width, height = plt_utils.scale(x, y)
        nx.draw_networkx(
            graph,
            pos=dict(zip(nodes, zip(x, y))),
            node_shape="H",
            node_size=50,
            node_color="#5a5b5b",
            edge_color="#C5C4C4",
            width=0.25,
            with_labels=False,
            ax=ax,
            arrows=False,
        )
        plt_utils.rm_spines(ax)

    def add_arrows(self):
        def arrow_between_axes(axA, axB):
            # Create the arrow
            # 1. Get transformation operators for axis and figure
            ax0tr = axA.transAxes  # Axis 0 -> Display
            ax1tr = axB.transAxes  # Axis 1 -> Display
            figtr = self.fig.transFigure.inverted()  # Display -> Figure
            # 2. Transform arrow start point from axis 0 to figure coordinates
            ptA = figtr.transform(ax0tr.transform((1, 0.5)))
            # 3. Transform arrow end point from axis 1 to figure coordinates
            ptB = figtr.transform(ax1tr.transform((0, 0.5)))
            # 4. Create the patch
            arrow = matplotlib.patches.FancyArrowPatch(
                ptA,
                ptB,
                transform=self.fig.transFigure,  # Place arrow in figure coord system
                # fc=self.fontcolor,
                # ec=self.fontcolor,
                #     connectionstyle="arc3",
                arrowstyle="simple, head_width=3, head_length=6, tail_width=0.15",
                alpha=1,
                mutation_scale=1.0,
            )
            arrow.set_lw(0.25)
            # 5. Add patch to list of objects to draw onto the figure
            self.fig.patches.append(arrow)

        if self.video and self.rendering:
            arrow_between_axes(self.ax_dict["video"], self.ax_dict["rendering"])
            arrow_between_axes(self.ax_dict["rendering"], self.ax_dict["retina box"])
        elif self.video:
            arrow_between_axes(self.ax_dict["video"], self.ax_dict["retina box"])
        elif self.rendering:
            arrow_between_axes(self.ax_dict["rendering"], self.ax_dict["retina box"])

        if self.motion_decoder and self.decoded_motion:
            arrow_between_axes(
                self.ax_dict["decoded box"], self.ax_dict["motion decoder"]
            )
            arrow_between_axes(
                self.ax_dict["motion decoder"], self.ax_dict["decoded motion"]
            )
        elif self.motion_decoder:
            arrow_between_axes(
                self.ax_dict["decoded box"], self.ax_dict["motion decoder"]
            )
        elif self.decoded_motion:
            arrow_between_axes(
                self.ax_dict["decoded box"], self.ax_dict["decoded motion"]
            )

    def add_cell_type_labels(self, fontsize=5):
        for label, ax in self.ax_dict.items():
            if label in self.cell_types:
                ax.annotate(
                    label,
                    (0, 0.9),
                    xycoords="axes fraction",
                    va="bottom",
                    ha="right",
                    fontsize=fontsize,
                )

    def add_neuropil_labels(self, fontsize=5):
        retina_cell_types = valfilter(lambda v: v == "retina", self.layout)
        axes = {
            cell_type: [ax for ax in self.axes if ax.get_label() == cell_type][0]
            for cell_type in retina_cell_types
        }
        (lefts, bottoms, rights, tops), _ = plt_utils.get_ax_positions(
            list(axes.values())
        )
        self.fig.text(
            lefts.min() + (rights.max() - lefts.min()) / 2,
            0,
            "retina",
            fontsize=fontsize,
            va="top",
            ha="center",
        )

        intermediate_cell_types = valfilter(lambda v: v == "intermediate", self.layout)
        axes = {
            cell_type: [ax for ax in self.axes if ax.get_label() == cell_type][0]
            for cell_type in intermediate_cell_types
        }
        (
            (lefts, bottoms, rights, tops),
            (
                centers,
                widths,
                height,
            ),
        ) = plt_utils.get_ax_positions(list(axes.values()))
        self.fig.text(
            lefts.min() + (rights.max() - lefts.min()) / 2,
            0,
            "lamina, medulla intrinsic cells, CT1",
            fontsize=fontsize,
            va="top",
            ha="center",
        )

        output_cell_types = valfilter(lambda v: v == "output", self.layout)
        axes = {
            cell_type: [ax for ax in self.axes if ax.get_label() == cell_type][0]
            for cell_type in output_cell_types
        }
        (
            (lefts, bottoms, rights, tops),
            (
                centers,
                widths,
                height,
            ),
        ) = plt_utils.get_ax_positions(list(axes.values()))
        self.fig.text(
            lefts.min() + (rights.max() - lefts.min()) / 2,
            0,
            "T-shaped, transmedullary cells",
            fontsize=fontsize,
            va="top",
            ha="center",
        )


def network_layout_axes(
    layout: Dict[str, str],
    cell_types: Optional[List[str]] = None,
    fig: Optional[plt.Figure] = None,
    figsize: List[int] = [16, 10],
    types_per_column: int = 8,
    region_spacing: int = 2,
    wspace: float = 0,
    hspace: float = 0,
    as_dict: bool = False,
    pos: Optional[Dict[str, List[float]]] = None,
) -> Tuple[
    plt.Figure, Union[List[plt.Axes], Dict[str, plt.Axes]], Dict[str, List[float]]
]:
    """
    Create axes for network layout.

    Args:
        layout: Dictionary mapping node types to layout positions.
        cell_types: List of cell types to include.
        fig: Existing figure to use.
        figsize: Size of the figure.
        types_per_column: Number of types per column.
        region_spacing: Spacing between regions.
        wspace: Width space between subplots.
        hspace: Height space between subplots.
        as_dict: Whether to return axes as a dictionary.
        pos: Pre-computed positions for nodes.

    Returns:
        Tuple containing the figure, axes, and node positions.
    """
    fig = fig or plt.figure(figsize=figsize)

    pos = pos or _network_graph_node_pos(
        layout, region_spacing=region_spacing, types_per_column=types_per_column
    )
    pos = {
        key: value
        for key, value in pos.items()
        if (cell_types is None or key in cell_types)
    }
    xy = np.array(list(pos.values()))
    # why pad this?
    # hpad = 0.05
    # wpad = 0.05
    hpad = 0.0
    wpad = 0.0
    fig, axes, xy_scaled = plt_utils.ax_scatter(
        xy[:, 0],
        xy[:, 1],
        fig=fig,
        wspace=wspace,
        hspace=hspace,
        hpad=hpad,
        wpad=wpad,
        alpha=0,
        labels=list(pos.keys()),
    )
    new_pos = {key: xy_scaled[i] for i, key in enumerate(pos.keys())}
    if as_dict:
        return (
            fig,
            {cell_type: axes[i] for i, cell_type in enumerate(new_pos)},
            new_pos,
        )
    return fig, axes, new_pos


def _network_graph_node_pos(
    layout: Dict[str, str], region_spacing: float = 2, types_per_column: int = 8
) -> Dict[str, List[float]]:
    """
    Compute (x, y) coordinates for nodes in a network graph.

    Args:
        layout: Dictionary mapping node types to layout positions.
        region_spacing: Spacing between regions.
        types_per_column: Number of types per column.

    Returns:
        Dictionary mapping node types to their (x, y) coordinates.

    Note:
        Special nodes like 'video', 'rendering', etc. are positioned at the middle
        y-coordinate of their respective columns.
    """
    x_coordinate = 0
    region_0 = "retina"
    pos = {}
    j = 0
    special_nodes = [
        "video",
        "rendering",
        "motion decoder",
        "decoded motion",
        "pixel-accurate motion",
    ]

    for typ in layout:
        if typ in special_nodes:
            region_spacing = 1.25
        if layout[typ] != region_0:
            x_coordinate += region_spacing
            j = 0
        elif (j % types_per_column) == 0 and j != 0:
            x_coordinate += 1
        y_coordinate = types_per_column - 1 - j % types_per_column
        pos[typ] = [x_coordinate, y_coordinate]
        region_0 = layout[typ]
        j += 1

    y_mid = (types_per_column - 1) / 2

    for node in special_nodes:
        if node in layout:
            pos[node][1] = y_mid

    if "pixel-accurate motion" in layout:
        pos["pixel-accurate motion"][1] = y_mid - 1.5

    return pos


### Overriding some networkx functions to make selfloop height and width controllable
class FancyArrowFactory:
    """Draw arrows with `matplotlib.patches.FancyarrowPatch`"""

    class ConnectionStyleFactory:
        def __init__(self, connectionstyles, selfloop_height, ax=None):
            import matplotlib as mpl
            import matplotlib.path  # call as mpl.path
            import numpy as np

            self.ax = ax
            self.mpl = mpl
            self.np = np
            self.base_connection_styles = [
                mpl.patches.ConnectionStyle(cs) for cs in connectionstyles
            ]
            self.n = len(self.base_connection_styles)
            self.selfloop_height = selfloop_height

        def curved(self, edge_index):
            return self.base_connection_styles[edge_index % self.n]

        def self_loop(
            self,
            edge_index,
            horizontal_shift=None,
            vertical_shift=None,
            origin_x=0,
            origin_y=0,
        ):
            def self_loop_connection(posA, posB, *args, **kwargs):
                if not self.np.all(posA == posB):
                    raise nx.NetworkXError(
                        "`self_loop` connection style method"
                        "is only to be used for self-loops"
                    )
                # this is called with _screen space_ values
                # so convert back to data space
                data_loc = self.ax.transData.inverted().transform(posA)
                v_shift = vertical_shift or 0.1 * self.selfloop_height
                h_shift = horizontal_shift or v_shift * 0.5
                # put the top of the loop first so arrow is not hidden by node
                path = self.np.asarray([
                    # 1
                    [origin_x, v_shift],
                    # 4 4 4
                    [h_shift, v_shift],
                    [h_shift, origin_y],
                    [origin_x, origin_y],
                    # 4 4 4
                    [-h_shift, origin_y],
                    [-h_shift, v_shift],
                    [origin_x, v_shift],
                ])
                # Rotate self loop 90 deg. if more than 1
                # This will allow for maximum of 4 visible self loops
                if edge_index % 4:
                    x, y = path.T
                    for _ in range(edge_index % 4):
                        x, y = y, -x
                    path = self.np.array([x, y]).T
                return self.mpl.path.Path(
                    self.ax.transData.transform(data_loc + path), [1, 4, 4, 4, 4, 4, 4]
                )

            return self_loop_connection

    def __init__(
        self,
        edge_pos,
        edgelist,
        nodelist,
        edge_indices,
        node_size,
        selfloop_height,
        connectionstyle="arc3",
        node_shape="o",
        arrowstyle="-",
        arrowsize=10,
        edge_color="k",
        alpha=None,
        linewidth=1.0,
        style="solid",
        min_source_margin=0,
        min_target_margin=0,
        ax=None,
    ):
        import matplotlib as mpl
        import matplotlib.patches  # call as mpl.patches
        import numpy as np

        if isinstance(connectionstyle, str):
            connectionstyle = [connectionstyle]
        elif np.iterable(connectionstyle):
            connectionstyle = list(connectionstyle)
        else:
            msg = "ConnectionStyleFactory arg `connectionstyle` must be str or iterable"
            raise nx.NetworkXError(msg)
        self.ax = ax
        self.mpl = mpl
        self.np = np
        self.edge_pos = edge_pos
        self.edgelist = edgelist
        self.nodelist = nodelist
        self.node_shape = node_shape
        self.min_source_margin = min_source_margin
        self.min_target_margin = min_target_margin
        self.edge_indices = edge_indices
        self.node_size = node_size
        self.connectionstyle_factory = self.ConnectionStyleFactory(
            connectionstyle, selfloop_height, ax
        )
        self.arrowstyle = arrowstyle
        self.arrowsize = arrowsize
        self.arrow_colors = mpl.colors.colorConverter.to_rgba_array(edge_color, alpha)
        self.linewidth = linewidth
        self.style = style
        if isinstance(arrowsize, list) and len(arrowsize) != len(edge_pos):
            raise ValueError("arrowsize should have the same length as edgelist")

    def __call__(self, i, h_shift=None, v_shift=None, x0=0, y0=0):
        (x1, y1), (x2, y2) = self.edge_pos[i]
        shrink_source = 0  # space from source to tail
        shrink_target = 0  # space from  head to target
        if self.np.iterable(self.node_size):  # many node sizes
            source, target = self.edgelist[i][:2]
            source_node_size = self.node_size[self.nodelist.index(source)]
            target_node_size = self.node_size[self.nodelist.index(target)]
            shrink_source = self.to_marker_edge(source_node_size, self.node_shape)
            shrink_target = self.to_marker_edge(target_node_size, self.node_shape)
        else:
            shrink_source = self.to_marker_edge(self.node_size, self.node_shape)
            shrink_target = shrink_source
        shrink_source = max(shrink_source, self.min_source_margin)
        shrink_target = max(shrink_target, self.min_target_margin)

        # scale factor of arrow head
        if isinstance(self.arrowsize, list):
            mutation_scale = self.arrowsize[i]
        else:
            mutation_scale = self.arrowsize

        if len(self.arrow_colors) > i:
            arrow_color = self.arrow_colors[i]
        elif len(self.arrow_colors) == 1:
            arrow_color = self.arrow_colors[0]
        else:  # Cycle through colors
            arrow_color = self.arrow_colors[i % len(self.arrow_colors)]

        if self.np.iterable(self.linewidth):
            if len(self.linewidth) > i:
                linewidth = self.linewidth[i]
            else:
                linewidth = self.linewidth[i % len(self.linewidth)]
        else:
            linewidth = self.linewidth

        if (
            self.np.iterable(self.style)
            and not isinstance(self.style, str)
            and not isinstance(self.style, tuple)
        ):
            if len(self.style) > i:
                linestyle = self.style[i]
            else:  # Cycle through styles
                linestyle = self.style[i % len(self.style)]
        else:
            linestyle = self.style

        if x1 == x2 and y1 == y2:
            connectionstyle = self.connectionstyle_factory.self_loop(
                self.edge_indices[i],
                horizontal_shift=h_shift,
                vertical_shift=v_shift,
                origin_x=x0,
                origin_y=y0,
            )
        else:
            connectionstyle = self.connectionstyle_factory.curved(self.edge_indices[i])
        return self.mpl.patches.FancyArrowPatch(
            (x1, y1),
            (x2, y2),
            arrowstyle=self.arrowstyle,
            shrinkA=shrink_source,
            shrinkB=shrink_target,
            mutation_scale=mutation_scale,
            color=arrow_color,
            linewidth=linewidth,
            connectionstyle=connectionstyle,
            linestyle=linestyle,
            zorder=1,  # arrows go behind nodes
        )

    def to_marker_edge(self, marker_size, marker):
        if marker in "s^>v<d":  # `large` markers need extra space
            return self.np.sqrt(2 * marker_size) / 2
        else:
            return self.np.sqrt(marker_size) / 2


def draw_networkx_edges(
    G,
    pos,
    edgelist=None,
    width=1.0,
    edge_color="k",
    style="solid",
    alpha=None,
    arrowstyle=None,
    arrowsize=10,
    edge_cmap=None,
    edge_vmin=None,
    edge_vmax=None,
    ax=None,
    arrows=None,
    label=None,
    node_size=300,
    nodelist=None,
    node_shape="o",
    connectionstyle="arc3",
    min_source_margin=0,
    min_target_margin=0,
    hide_ticks=True,
    selfloop_x0=0,
    selfloop_y0=0,
    selfloop_h_shift=None,
    selfloop_v_shift=None,
):
    r"""Draw the edges of the graph G.

    This draws only the edges of the graph G.

    Parameters
    ----------
    G : graph
        A networkx graph

    pos : dictionary
        A dictionary with nodes as keys and positions as values.
        Positions should be sequences of length 2.

    edgelist : collection of edge tuples (default=G.edges())
        Draw only specified edges

    width : float or array of floats (default=1.0)
        Line width of edges

    edge_color : color or array of colors (default='k')
        Edge color. Can be a single color or a sequence of colors with the same
        length as edgelist. Color can be string or rgb (or rgba) tuple of
        floats from 0-1. If numeric values are specified they will be
        mapped to colors using the edge_cmap and edge_vmin,edge_vmax parameters.

    style : string or array of strings (default='solid')
        Edge line style e.g.: '-', '--', '-.', ':'
        or words like 'solid' or 'dashed'.
        Can be a single style or a sequence of styles with the same
        length as the edge list.
        If less styles than edges are given the styles will cycle.
        If more styles than edges are given the styles will be used sequentially
        and not be exhausted.
        Also, `(offset, onoffseq)` tuples can be used as style instead of a strings.
        (See `matplotlib.patches.FancyArrowPatch`: `linestyle`)

    alpha : float or array of floats (default=None)
        The edge transparency.  This can be a single alpha value,
        in which case it will be applied to all specified edges. Otherwise,
        if it is an array, the elements of alpha will be applied to the colors
        in order (cycling through alpha multiple times if necessary).

    edge_cmap : Matplotlib colormap, optional
        Colormap for mapping intensities of edges

    edge_vmin,edge_vmax : floats, optional
        Minimum and maximum for edge colormap scaling

    ax : Matplotlib Axes object, optional
        Draw the graph in the specified Matplotlib axes.

    arrows : bool or None, optional (default=None)
        If `None`, directed graphs draw arrowheads with
        `~matplotlib.patches.FancyArrowPatch`, while undirected graphs draw edges
        via `~matplotlib.collections.LineCollection` for speed.
        If `True`, draw arrowheads with FancyArrowPatches (bendable and stylish).
        If `False`, draw edges using LineCollection (linear and fast).

        Note: Arrowheads will be the same color as edges.

    arrowstyle : str (default='-\|>' for directed graphs)
        For directed graphs and `arrows==True` defaults to '-\|>',
        For undirected graphs default to '-'.

        See `matplotlib.patches.ArrowStyle` for more options.

    arrowsize : int (default=10)
        For directed graphs, choose the size of the arrow head's length and
        width. See `matplotlib.patches.FancyArrowPatch` for attribute
        `mutation_scale` for more info.

    connectionstyle : string or iterable of strings (default="arc3")
        Pass the connectionstyle parameter to create curved arc of rounding
        radius rad. For example, connectionstyle='arc3,rad=0.2'.
        See `matplotlib.patches.ConnectionStyle` and
        `matplotlib.patches.FancyArrowPatch` for more info.
        If Iterable, index indicates i'th edge key of MultiGraph

    node_size : scalar or array (default=300)
        Size of nodes. Though the nodes are not drawn with this function, the
        node size is used in determining edge positioning.

    nodelist : list, optional (default=G.nodes())
       This provides the node order for the `node_size` array (if it is an array).

    node_shape :  string (default='o')
        The marker used for nodes, used in determining edge positioning.
        Specification is as a `matplotlib.markers` marker, e.g. one of 'so^>v<dph8'.

    label : None or string
        Label for legend

    min_source_margin : int (default=0)
        The minimum margin (gap) at the beginning of the edge at the source.

    min_target_margin : int (default=0)
        The minimum margin (gap) at the end of the edge at the target.

    hide_ticks : bool, optional
        Hide ticks of axes. When `True` (the default), ticks and ticklabels
        are removed from the axes. To set ticks and tick labels to the pyplot default,
        use ``hide_ticks=False``.

    Returns
    -------
     matplotlib.collections.LineCollection or a list of matplotlib.patches.FancyArrowPatch
        If ``arrows=True``, a list of FancyArrowPatches is returned.
        If ``arrows=False``, a LineCollection is returned.
        If ``arrows=None`` (the default), then a LineCollection is returned if
        `G` is undirected, otherwise returns a list of FancyArrowPatches.

    Notes
    -----
    For directed graphs, arrows are drawn at the head end.  Arrows can be
    turned off with keyword arrows=False or by passing an arrowstyle without
    an arrow on the end.

    Be sure to include `node_size` as a keyword argument; arrows are
    drawn considering the size of nodes.

    Self-loops are always drawn with `~matplotlib.patches.FancyArrowPatch`
    regardless of the value of `arrows` or whether `G` is directed.
    When ``arrows=False`` or ``arrows=None`` and `G` is undirected, the
    FancyArrowPatches corresponding to the self-loops are not explicitly
    returned. They should instead be accessed via the ``Axes.patches``
    attribute (see examples).

    Examples
    --------
    >>> G = nx.dodecahedral_graph()
    >>> edges = nx.draw_networkx_edges(G, pos=nx.spring_layout(G))

    >>> G = nx.DiGraph()
    >>> G.add_edges_from([(1, 2), (1, 3), (2, 3)])
    >>> arcs = nx.draw_networkx_edges(G, pos=nx.spring_layout(G))
    >>> alphas = [0.3, 0.4, 0.5]
    >>> for i, arc in enumerate(arcs):  # change alpha values of arcs
    ...     arc.set_alpha(alphas[i])

    The FancyArrowPatches corresponding to self-loops are not always
    returned, but can always be accessed via the ``patches`` attribute of the
    `matplotlib.Axes` object.

    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> G = nx.Graph([(0, 1), (0, 0)])  # Self-loop at node 0
    >>> edge_collection = nx.draw_networkx_edges(G, pos=nx.circular_layout(G), ax=ax)
    >>> self_loop_fap = ax.patches[0]

    Also see the NetworkX drawing examples at
    https://networkx.org/documentation/latest/auto_examples/index.html

    See Also
    --------
    draw
    draw_networkx
    draw_networkx_nodes
    draw_networkx_labels
    draw_networkx_edge_labels

    """
    import warnings

    import matplotlib as mpl
    import matplotlib.collections  # call as mpl.collections
    import matplotlib.colors  # call as mpl.colors
    import matplotlib.pyplot as plt
    import numpy as np

    # The default behavior is to use LineCollection to draw edges for
    # undirected graphs (for performance reasons) and use FancyArrowPatches
    # for directed graphs.
    # The `arrows` keyword can be used to override the default behavior
    if arrows is None:
        use_linecollection = not (G.is_directed() or G.is_multigraph())
    else:
        if not isinstance(arrows, bool):
            raise TypeError("Argument `arrows` must be of type bool or None")
        use_linecollection = not arrows

    if isinstance(connectionstyle, str):
        connectionstyle = [connectionstyle]
    elif np.iterable(connectionstyle):
        connectionstyle = list(connectionstyle)
    else:
        msg = "draw_networkx_edges arg `connectionstyle` must be str or iterable"
        raise nx.NetworkXError(msg)

    # Some kwargs only apply to FancyArrowPatches. Warn users when they use
    # non-default values for these kwargs when LineCollection is being used
    # instead of silently ignoring the specified option
    if use_linecollection:
        msg = (
            "\n\nThe {0} keyword argument is not applicable when drawing edges\n"
            "with LineCollection.\n\n"
            "To make this warning go away, either specify `arrows=True` to\n"
            "force FancyArrowPatches or use the default values.\n"
            "Note that using FancyArrowPatches may be slow for large graphs.\n"
        )
        if arrowstyle is not None:
            warnings.warn(msg.format("arrowstyle"), category=UserWarning, stacklevel=2)
        if arrowsize != 10:
            warnings.warn(msg.format("arrowsize"), category=UserWarning, stacklevel=2)
        if min_source_margin != 0:
            warnings.warn(
                msg.format("min_source_margin"), category=UserWarning, stacklevel=2
            )
        if min_target_margin != 0:
            warnings.warn(
                msg.format("min_target_margin"), category=UserWarning, stacklevel=2
            )
        if any(cs != "arc3" for cs in connectionstyle):
            warnings.warn(
                msg.format("connectionstyle"), category=UserWarning, stacklevel=2
            )

    # NOTE: Arrowstyle modification must occur after the warnings section
    if arrowstyle is None:
        arrowstyle = "-|>" if G.is_directed() else "-"

    if ax is None:
        ax = plt.gca()

    if edgelist is None:
        edgelist = list(G.edges)  # (u, v, k) for multigraph (u, v) otherwise

    if len(edgelist):
        if G.is_multigraph():
            key_count = collections.defaultdict(lambda: itertools.count(0))
            edge_indices = [next(key_count[tuple(e[:2])]) for e in edgelist]
        else:
            edge_indices = [0] * len(edgelist)
    else:  # no edges!
        return []

    if nodelist is None:
        nodelist = list(G.nodes())

    # FancyArrowPatch handles color=None different from LineCollection
    if edge_color is None:
        edge_color = "k"

    # set edge positions
    edge_pos = np.asarray([(pos[e[0]], pos[e[1]]) for e in edgelist])

    # Check if edge_color is an array of floats and map to edge_cmap.
    # This is the only case handled differently from matplotlib
    if (
        np.iterable(edge_color)
        and (len(edge_color) == len(edge_pos))
        and np.all([isinstance(c, Number) for c in edge_color])
    ):
        if edge_cmap is not None:
            assert isinstance(edge_cmap, mpl.colors.Colormap)
        else:
            edge_cmap = plt.get_cmap()
        if edge_vmin is None:
            edge_vmin = min(edge_color)
        if edge_vmax is None:
            edge_vmax = max(edge_color)
        color_normal = mpl.colors.Normalize(vmin=edge_vmin, vmax=edge_vmax)
        edge_color = [edge_cmap(color_normal(e)) for e in edge_color]

    # compute initial view
    minx = np.amin(np.ravel(edge_pos[:, :, 0]))
    maxx = np.amax(np.ravel(edge_pos[:, :, 0]))
    miny = np.amin(np.ravel(edge_pos[:, :, 1]))
    maxy = np.amax(np.ravel(edge_pos[:, :, 1]))
    w = maxx - minx
    h = maxy - miny

    # Self-loops are scaled by view extent, except in cases the extent
    # is 0, e.g. for a single node. In this case, fall back to scaling
    # by the maximum node size
    selfloop_height = h if h != 0 else 0.005 * np.array(node_size).max()
    fancy_arrow_factory = FancyArrowFactory(
        edge_pos,
        edgelist,
        nodelist,
        edge_indices,
        node_size,
        selfloop_height,
        connectionstyle,
        node_shape,
        arrowstyle,
        arrowsize,
        edge_color,
        alpha,
        width,
        style,
        min_source_margin,
        min_target_margin,
        ax=ax,
    )

    # Draw the edges
    if use_linecollection:
        edge_collection = mpl.collections.LineCollection(
            edge_pos,
            colors=edge_color,
            linewidths=width,
            antialiaseds=(1,),
            linestyle=style,
            alpha=alpha,
        )
        edge_collection.set_cmap(edge_cmap)
        edge_collection.set_clim(edge_vmin, edge_vmax)
        edge_collection.set_zorder(1)  # edges go behind nodes
        edge_collection.set_label(label)
        ax.add_collection(edge_collection)
        edge_viz_obj = edge_collection

        # Make sure selfloop edges are also drawn
        # ---------------------------------------
        selfloops_to_draw = [loop for loop in nx.selfloop_edges(G) if loop in edgelist]
        if selfloops_to_draw:
            edgelist_tuple = list(map(tuple, edgelist))
            arrow_collection = []
            for loop in selfloops_to_draw:
                i = edgelist_tuple.index(loop)
                arrow = fancy_arrow_factory(
                    i,
                    h_shift=selfloop_h_shift,
                    v_shift=selfloop_v_shift,
                    x0=selfloop_x0,
                    y0=selfloop_y0,
                )
                arrow_collection.append(arrow)
                ax.add_patch(arrow)
    else:
        edge_viz_obj = []
        for i in range(len(edgelist)):
            arrow = fancy_arrow_factory(
                i,
                h_shift=selfloop_h_shift,
                v_shift=selfloop_v_shift,
                x0=selfloop_x0,
                y0=selfloop_y0,
            )
            ax.add_patch(arrow)
            edge_viz_obj.append(arrow)

    # update view after drawing
    padx, pady = 0.05 * w, 0.05 * h
    corners = (minx - padx, miny - pady), (maxx + padx, maxy + pady)
    ax.update_datalim(corners)
    ax.autoscale_view()

    if hide_ticks:
        ax.tick_params(
            axis="both",
            which="both",
            bottom=False,
            left=False,
            labelbottom=False,
            labelleft=False,
        )

    return edge_viz_obj



import os
import tempfile
from pathlib import Path

import flyvis


def test_default_root_dir():
    """Test that the default root directory is correctly set relative to the package."""
    # Clear any existing FLYVIS_ROOT_DIR environment variable
    if "FLYVIS_ROOT_DIR" in os.environ:
        del os.environ["FLYVIS_ROOT_DIR"]

    # Get the resolved root directory
    root_dir = flyvis.resolve_root_dir()

    # Expected default is repo_dir/data
    expected_dir = flyvis.repo_dir / "data"
    assert root_dir == expected_dir.absolute()


def test_env_var_root_dir():
    """Test that FLYVIS_ROOT_DIR environment variable is respected."""
    with tempfile.TemporaryDirectory() as temp_dir:
        # Set environment variable to temp directory
        os.environ["FLYVIS_ROOT_DIR"] = temp_dir

        # Get the resolved root directory
        root_dir = flyvis.resolve_root_dir()

        assert root_dir == Path(temp_dir).absolute()


def test_dotenv_file():
    """Test that .env file is properly loaded."""
    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        # Create a .env file
        env_content = f'FLYVIS_ROOT_DIR={temp_dir}'
        with open(temp_path / '.env', 'w') as f:
            f.write(env_content)

        # Clear any existing environment variable
        if "FLYVIS_ROOT_DIR" in os.environ:
            del os.environ["FLYVIS_ROOT_DIR"]

        # Change to temp directory and reload environment
        original_dir = os.getcwd()
        os.chdir(temp_dir)

        try:
            import dotenv

            dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))

            # Get the resolved root directory
            root_dir = flyvis.resolve_root_dir()

            assert root_dir == temp_path.absolute()

        finally:
            # Restore original directory
            os.chdir(original_dir)


def test_path_expansion():
    """Test that user path expansion works correctly."""
    test_path = "~/flyvis_test"
    os.environ["FLYVIS_ROOT_DIR"] = test_path

    root_dir = flyvis.resolve_root_dir()
    expected_dir = Path(test_path).expanduser().absolute()

    assert root_dir == expected_dir



from __future__ import annotations

import contextlib
import operator
import warnings
from functools import wraps
from typing import Iterable, List

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr


def where_xarray(
    dataset: xr.Dataset | xr.DataArray,
    rtol: float = 1.0e-5,
    atol: float = 1.0e-8,
    **kwargs,
) -> xr.Dataset | xr.DataArray:
    """Return a subset of the xarray Dataset or DataArray where coordinates meet
    specified query-like conditions.

    Args:
        dataset: The dataset or data array to filter.
        rtol: Relative tolerance for floating point comparisons.
        atol: Absolute tolerance for floating point comparisons.
        **kwargs: Query-like conditions on coordinates. Conditions can be specified as:
            - Strings with comma-separated conditions (interpreted as AND).
            - Iterables (lists, tuples) representing multiple conditions
                (interpreted as OR).
            - Single values for equality conditions.

    Returns:
        The filtered dataset or data array.

    Example:
        ```python
        filtered_ds = where_xarray(
            ds,
            cell_type=["T4a", "T4b"],
            time="<1.0,>0",
            intensity=1.0,
            radius=6,
            width=2.4
        )
        ```
    """
    # Force evaluation of coordinates
    # Heisenbug, strangely required for the where() method to work
    # to circumvent AttributeError: 'ScipyArrayWrapper' object has no attribute 'oindex'
    for _, coord in dataset.coords.items():
        _ = coord.values.dtype

    # Define a mapping of operators from string to functions
    operators = {
        '>=': operator.ge,
        '<=': operator.le,
        '==': operator.eq,
        '!=': operator.ne,
        '>': operator.gt,
        '<': operator.lt,
    }

    # Sort operators by length in descending order to match multi-character operators
    # first
    sorted_operators = sorted(operators.keys(), key=len, reverse=True)

    def parse_condition(cond_str):
        """Parse a single condition string into (operator_function, target_value)."""
        cond_str = cond_str.strip()
        for op_str in sorted_operators:
            if cond_str.startswith(op_str):
                target = cond_str[len(op_str) :].strip()
                with contextlib.suppress(ValueError):
                    target = float(target)
                return (operators[op_str], target)
        # If no operator is found, assume equality
        try:
            target = float(cond_str)
        except ValueError:
            target = cond_str
        return (operator.eq, target)

    filtered_dataset = dataset

    for coord_name, condition in kwargs.items():
        # Check if coord_name is a coordinate in the dataset
        if coord_name not in dataset.coords:
            raise ValueError(f"Coordinate '{coord_name}' not found in the dataset.")

        coord_values = dataset.coords[coord_name]
        coord_mask = xr.ones_like(coord_values, dtype=bool)  # Initialize mask as all True

        if isinstance(condition, str):
            # String conditions: multiple conditions separated by commas (AND logic)
            condition_strings = [c.strip() for c in condition.split(',') if c.strip()]
            for cond_str in condition_strings:
                op_func, target_value = parse_condition(cond_str)

                if np.issubdtype(coord_values.dtype, np.floating):
                    if op_func == operator.eq:
                        mask = np.isclose(
                            coord_values, target_value, atol=atol, rtol=rtol
                        )
                    else:
                        mask = op_func(coord_values, target_value)
                else:
                    mask = op_func(coord_values, target_value)

                # Combine masks using logical AND
                coord_mask &= xr.DataArray(
                    mask, dims=coord_values.dims, coords=coord_values.coords
                )

        elif isinstance(condition, Iterable) and not isinstance(condition, (str, bytes)):
            # Iterable conditions: each element is a separate condition (OR logic)
            temp_mask = xr.zeros_like(
                coord_values, dtype=bool
            )  # Initialize mask as all False
            for item in condition:
                if isinstance(item, str):
                    # Parse condition string
                    op_func, target_value = parse_condition(item)
                else:
                    # Assume equality if not a string condition
                    op_func, target_value = operator.eq, item

                if np.issubdtype(coord_values.dtype, np.floating):
                    if op_func == operator.eq:
                        mask = np.isclose(
                            coord_values, target_value, atol=atol, rtol=rtol
                        )
                    else:
                        mask = op_func(coord_values, target_value)
                else:
                    mask = op_func(coord_values, target_value)

                # Combine masks using logical OR
                temp_mask |= xr.DataArray(
                    mask, dims=coord_values.dims, coords=coord_values.coords
                )
            coord_mask &= temp_mask  # Apply OR mask with existing mask
        else:
            # Single non-string, non-iterable value: assume equality
            op_func, target_value = operator.eq, condition
            if np.issubdtype(coord_values.dtype, np.floating):
                if op_func == operator.eq:
                    mask = np.isclose(coord_values, target_value, atol=atol, rtol=rtol)
                else:
                    mask = op_func(coord_values, target_value)
            else:
                mask = op_func(coord_values, target_value)
            coord_mask &= xr.DataArray(
                mask, dims=coord_values.dims, coords=coord_values.coords
            )

        # Apply the combined mask
        filtered_dataset = filtered_dataset.where(coord_mask, drop=True)

    return filtered_dataset


def plot_traces(
    dataset: xr.DataArray | xr.Dataset,
    key: str,
    x: str,
    legend_labels: List[str] = [],
    extra_legend_coords: List[str] = [],
    plot_kwargs: dict = {},
    **kwargs,
) -> plt.Axes:
    """Plot the flash response traces from the dataset, optionally filtered by various
    parameters.

    Args:
        dataset: The dataset containing the responses to plot.
        key: The key of the data to plot if the dataset is a Dataset.
        x: The dimension to use as the x-axis.
        legend_labels: List of coordinates to include in the legend.
        extra_legend_coords: Additional coordinates to include in the legend.
        plot_kwargs: Additional keyword arguments to pass to the plot function.
        **kwargs: Query-like conditions on coordinates.

    Returns:
        The matplotlib axes object containing the plot.

    Note:
        Query-like conditions can be specified as:

        - Strings with comma-separated conditions (e.g., time='<0.5,>0.1')
        - Lists for equality conditions (e.g., cell_type=["T4a", "T4b"])
        - Single values for equality conditions (e.g., intensity=1.0)
    """
    traces = dataset.custom.where(**kwargs)

    if key:
        traces = traces[key]

    arg_df = traces.sample.to_dataframe()

    # Stack all dims besides x
    stack_dims = [dim for dim in traces.dims if dim not in list(traces[x].coords.keys())]
    # logging.info("Stacking dimensions: %s", stack_dims)
    traces = traces.stack(traces=stack_dims)

    num_stacks = traces.sizes.get('traces', 0)
    if num_stacks > 250:
        warnings.warn(
            f"The traces stack has {num_stacks} elements.",
            UserWarning,
            stacklevel=2,
        )

    original_legend_labels = [col for col in arg_df.columns if col != 'sample']
    if x in original_legend_labels:
        # cannot set legend for x-axis values
        original_legend_labels = []

    stacked_legend_labels = list(stack_dims)

    legend_labels = (
        legend_labels
        or stacked_legend_labels + extra_legend_coords + original_legend_labels
    )
    legend_table = [np.atleast_1d(traces[col].data) for col in legend_labels]

    # Confirm all elements are 1D arrays of equal length
    try:
        legend_table = np.column_stack(legend_table)
    except ValueError as e:
        raise ValueError(
            "All elements in legend_coords must be 1D arrays of equal length. "
            "Specify legend_labels to use only a subset of the coordinates."
        ) from e

    legend_info = np.array([
        ", ".join([f"{col}: {value}" for col, value in zip(legend_labels, row)])
        for row in legend_table
    ])
    traces = traces.assign_coords(legend_info=("traces", legend_info))

    traces.plot.line(x=x, hue="legend_info", **plot_kwargs)

    ax = plt.gca()

    legend = ax.get_legend()
    if legend is not None:
        legend.set_title(None)

    return ax


class CustomAccessor:
    """Custom accessor for xarray objects providing additional functionality.

    Attributes:
        _obj: The xarray object being accessed.
    """

    def __init__(self, xarray_obj: xr.Dataset | xr.DataArray):
        self._obj = xarray_obj

    @wraps(where_xarray)
    def where(self, **kwargs) -> xr.Dataset | xr.DataArray:
        return where_xarray(self._obj, **kwargs)

    @wraps(plot_traces)
    def plot_traces(
        self,
        x: str,
        key: str = "",
        legend_labels: List[str] = [],
        extra_legend_coords: List[str] = [],
        plot_kwargs: dict = {},
        **kwargs,
    ) -> plt.Axes:
        """Plot traces from the xarray object.

        Args:
            x: The dimension to use as the x-axis.
            key: The key of the data to plot if the dataset is a Dataset.
            legend_labels: List of coordinates to include in the legend.
            extra_legend_coords: Additional coordinates to include in the legend.
            plot_kwargs: Additional keyword arguments to pass to the plot function.
            **kwargs: Query-like conditions on coordinates.

        Returns:
            The matplotlib axes object containing the plot.

        Example:
            Overlay stimulus and response traces:
            ```python
            fig, ax = plt.subplots()
            r.custom.plot_traces(
                key='stimulus',
                x='time',
                speed=[19, 25],
                intensity=1,
                angle=90,
                u_in=0,
                v_in=0,
                plot_kwargs=dict(ax=ax),
                time='>0,<1.0'
            )
            r.custom.plot_traces(
                key='responses',
                x='time',
                speed=[19, 25],
                cell_type='T4c',
                intensity=1,
                angle=90,
                network_id=0,
                plot_kwargs=dict(ax=ax),
                time='>0,<1.0'
            )
            ```

            Polar plot:
            ```python
            prs = peak_responses(stims_and_resps_moving_edges).custom.where(
                cell_type="T4c",
                intensity=1,
                speed=19,
            )
            prs['angle'] = np.radians(prs.angle)
            ax = plt.subplots(subplot_kw={"projection": "polar"})[1]
            prs.custom.plot_traces(
                x="angle",
                legend_labels=["network_id"],
                plot_kwargs={"add_legend": False, "ax": ax, "color": "b"},
            )
            ```
        """
        return plot_traces(
            self._obj,
            key,
            x,
            legend_labels,
            extra_legend_coords,
            plot_kwargs,
            **kwargs,
        )



import re
from typing import Iterable, List, Optional, Tuple, Union

import numpy as np
import torch
from numpy.typing import NDArray

from flyvis import connectome
from flyvis.utils import groundtruth_utils


def order_node_type_list(
    node_types: List[str],
    groups: List[str] = [
        r"R\d",
        r"L\d",
        r"Lawf\d",
        r"A",
        r"C\d",
        r"CT\d.*",
        r"Mi\d{1,2}",
        r"T\d{1,2}.*",
        r"Tm.*\d{1,2}.*",
    ],
) -> Tuple[List[str], List[int]]:
    """Orders a list of node types by the regular expressions defined in groups.

    Args:
        node_types: Messy list of nodes.
        groups: Ordered list of regular expressions to sort node_types.

    Returns:
        A tuple containing:
        - Ordered node type list
        - Corresponding sorting indices

    Raises:
        AssertionError: If sorting doesn't include all cell types.
        ValueError: If sorting fails due to length mismatch.
    """
    if node_types is None:
        return None, None

    _len = len(node_types)

    def sort_numeric(string):
        """Used in sorted(list, key=sort_fn) for sorting
        lists including 0 to 4 digits after the character.
        """
        regular_expression = r"\d+"
        match = re.search(regular_expression, string)
        if not match:
            # For example Am types are not numbered.
            return string
        return re.sub(regular_expression, f"{int(match.group()):04}", string)

    #     breakpoint()
    type_groups = {index: [] for index in range(len(groups))}
    type_groups.update({len(groups) + 1: []})  # for unmatched types.
    matched = {cell_type: False for cell_type in node_types}
    for node_index, cell_type in enumerate(node_types):
        for group_index, regular_expression in enumerate(groups):
            if re.match(regular_expression, cell_type):
                type_groups[group_index].append((node_index, cell_type))
                matched[cell_type] = True
        if matched[cell_type]:
            pass
        else:
            type_groups[len(groups) + 1].append((node_index, cell_type))

    # ordered = [y for x in type_groups.values() for y in sorted(x, key=lambda z:
    # sort_fn(z[1]))]
    ordered = []
    for x in type_groups.values():
        for y in sorted(x, key=lambda z: sort_numeric(z[1])):
            ordered.append(y)
    index = [y[0] for y in ordered]
    nodes = [y[1] for y in ordered]

    if set(node_types) - set(nodes):
        print(set(node_types) - set(nodes))
        raise AssertionError(
            "Defined sorting through regular expressions does not include all cell"
            " types."
        )

    if _len != len(nodes) or _len != len(index):
        raise ValueError(
            "sorting failed because the resulting array if of " " different length"
        )

    return nodes, index


def get_index_mapping_lists(from_list: List[str], to_list: List[str]) -> List[int]:
    """Get indices to sort and filter from_list by occurrence of items in to_list.

    The indices are useful to sort or filter another list or tensor that
    is an ordered mapping to items in from_list to the order of items in to_list.

    Args:
        from_list: Original list of items.
        to_list: Target list of items.

    Returns:
        List of indices for sorting.

    Example:
        ```python
        from_list = ["a", "b", "c"]
        mapping_to_from_list = [1, 2, 3]
        to_list = ["c", "a", "b"]
        sort_index = get_index_mapping_lists(from_list, to_list)
        sorted_list = [mapping_to_from_list[i] for i in sort_index]
        # sorted_list will be [3, 1, 2]
        ```
    """
    if isinstance(from_list, np.ndarray):
        from_list = from_list.tolist()
    if isinstance(to_list, np.ndarray):
        to_list = to_list.tolist()
    return [from_list.index(item) for item in to_list]


def sort_by_mapping_lists(
    from_list: List[str],
    to_list: List[str],
    tensor: Union[np.ndarray, torch.Tensor],
    axis: int = 0,
) -> np.ndarray:
    """Sort and filter a tensor along an axis indexed by from_list to match to_list.

    Args:
        from_list: Original list of items.
        to_list: Target list of items.
        tensor: Tensor to be sorted.
        axis: Axis along which to sort the tensor.

    Returns:
        Sorted numpy array.
    """
    tensor = np.array(tensor)
    if axis != 0:
        tensor = np.transpose(tensor, axes=(axis, 0))
    sort_index = get_index_mapping_lists(from_list, to_list)
    tensor = np.array([tensor[i] for i in sort_index])
    if axis != 0:
        tensor = np.transpose(tensor, axes=(axis, 0))
    return tensor


def nodes_list_sorting_on_off_unknown(
    cell_types: Optional[List[str]] = None,
) -> List[str]:
    """Sort node list based on on/off/unknown polarity.

    Args:
        cell_types: List of cell types to sort. If None, uses all types from
                    groundtruth_utils.polarity.

    Returns:
        Sorted list of cell types.
    """
    value = {1: 1, -1: 2, 0: 3}
    preferred_contrasts = groundtruth_utils.polarity
    cell_types = list(preferred_contrasts) if cell_types is None else cell_types
    preferred_contrasts = {
        k: value[v] for k, v in preferred_contrasts.items() if k in cell_types
    }
    preferred_contrasts = dict(sorted(preferred_contrasts.items(), key=lambda k: k[1]))
    nodes_list = list(preferred_contrasts.keys())
    return nodes_list


class NodeIndexer(dict):
    """Attribute-style accessible map from cell types to indices.

    Args:
        connectome: Connectome object. The cell types are taken from the
            connectome and references are created in order.
        unique_cell_types: Array of unique cell types. Optional.
            To specify the mapping from cell types to indices in provided order.

    Attributes:
        unique_cell_types (NDArray[str]): Array of unique cell types.
        central_cells_index (Optional[NDArray[int]]): Array of indices of central cells.

    Raises:
        ValueError: If neither connectome nor unique_cell_types is provided.
    """

    def __init__(
        self,
        connectome: Optional["connectome.ConnectomeFromAvgFilters"] = None,
        unique_cell_types: Optional[NDArray[str]] = None,
    ):
        # if connectome is specified, the indices are taken from the connectome
        # and reference to positions in the entire list of nodes/cells
        if connectome is not None and unique_cell_types is None:
            self.unique_cell_types = connectome.unique_cell_types[:].astype("str")
            self.central_cells_index = connectome.central_cells_index[:]
        # alternatively the mapping can be specified from a list of cell types
        # and reference to positions in order of the list
        elif connectome is None and unique_cell_types is not None:
            self.unique_cell_types = unique_cell_types
            self.central_cells_index = None
        else:
            raise ValueError("either cell types or connectome must be specified")
        for index, cell_type in enumerate(self.unique_cell_types):
            super().__setitem__(cell_type, index)

    def __dir__(self):
        return list(set([*dict.__dir__(self), *dict.__iter__(self)]))

    def __len__(self):
        return len(self.unique_cell_types)

    def __iter__(self):
        for cell_type in self.unique_cell_types:
            yield cell_type

    def __getattr__(self, key):
        if isinstance(key, str):
            pass
        elif isinstance(key, Iterable):
            return [dict.__getitem__(self, _key) for _key in key]
        return dict.__getitem__(self, key)

    def __getitem__(self, key):
        return self.__getattr__(key)


class CellTypeArray:
    """Attribute-style accessible map from cell types to coordinates in array.

    Args:
        array: Has the dim-th axis corresponding to unique cell types
            in the connectome or provided cell types.
        connectome: Connectome object.
        cell_types: Array of cell types.
        dim: Axis corresponding to unique cell types.

    Attributes:
        node_indexer (NodeIndexer): Indexer for cell types.
        array (NDArray): The array of cell type data.
        dim (int): Dimension corresponding to cell types.
        cell_types (NDArray[str]): Array of unique cell types.
    """

    node_indexer: NodeIndexer = None
    array: NDArray = None
    dim: float = None

    def __init__(
        self,
        array: Union[NDArray, torch.Tensor],
        connectome: Optional["connectome.ConnectomeFromAvgFilters"] = None,
        cell_types: Optional[NDArray[str]] = None,
        dim: int = -1,
    ):
        self.array = array
        self.dim = dim
        self.node_indexer = NodeIndexer(connectome, cell_types)
        self.cell_types = self.node_indexer.unique_cell_types

    def __bool__(self):
        return self.array is not None

    def __iter__(self):
        for cell_type in self.node_indexer.unique_cell_types:
            yield cell_type

    def __dir__(self):
        return list(
            set([
                *object.__dir__(self),
                *dict.__dir__(self.node_indexer),
                *dict.__iter__(self.node_indexer),
            ])
        )

    @property
    def shape(self):
        if self.array is not None:
            return self.array.shape
        return []

    def __repr__(self):
        shape = list(self.shape)
        desc = f"Array({tuple(shape)})"
        return {k: desc for k in self}.__repr__()

    def values(self):
        return [self[k] for k in self]

    def keys(self):
        return [k for k in self]

    def items(self):
        return [(k, self[k]) for k in self]

    def __len__(self):
        return len(self.node_indexer.unique_cell_types)

    def __getattr__(self, key):
        if self.node_indexer is not None:
            if isinstance(key, slice) and key == slice(None):
                return self.array
            elif isinstance(key, str) and key in self.node_indexer.unique_cell_types:
                indices = np.int_([dict.__getitem__(self.node_indexer, key)])
            elif isinstance(key, Iterable) and all([
                _key in self.node_indexer.unique_cell_types for _key in key
            ]):
                indices = np.int_([
                    dict.__getitem__(self.node_indexer, _key) for _key in key
                ])
            elif key in self.node_indexer.__dir__():
                return object.__getattribute__(self.node_indexer, key)
            else:
                return object.__getattribute__(self, key)
            return np.take(self.array, indices, axis=self.dim)
        return object.__getattribute__(self, key)

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __setitem__(self, key, value):
        if self.node_indexer is not None and key in self.node_indexer.unique_cell_types:
            if value.shape[-1] != 1:
                value = np.expand_dims(value, self.dim)
            if self.array is None:
                n_cell_types = len(self.node_indexer.unique_cell_types)
                shape = list(value.shape)
                shape[self.dim] = n_cell_types
                self.array = np.zeros(shape)
            # breakpoint()
            index = dict.__getitem__(self.node_indexer, key)
            np.put_along_axis(
                self.array,
                np.expand_dims(np.array([index]), list(range(len(self.array.shape[1:])))),
                value,
                self.dim,
            )
        else:
            object.__setattr__(self, key, value)

    def __setattr__(self, key, value):
        return self.__setitem__(key, value)

    def from_cell_types(self, cell_types):
        activity = self[cell_types]
        return CellTypeArray(
            activity,
            cell_types=cell_types,
            dim=self.dim,
        )


# cell type layout for visualization
layout = {
    "R1": "retina",
    "R2": "retina",
    "R3": "retina",
    "R4": "retina",
    "R5": "retina",
    "R6": "retina",
    "R7": "retina",
    "R8": "retina",
    "L1": "intermediate",
    "L2": "intermediate",
    "L3": "intermediate",
    "L4": "intermediate",
    "L5": "intermediate",
    "Lawf1": "intermediate",
    "Lawf2": "intermediate",
    "Am": "intermediate",
    "C2": "intermediate",
    "C3": "intermediate",
    "CT1(Lo1)": "intermediate",
    "CT1(M10)": "intermediate",
    "Mi1": "intermediate",
    "Mi2": "intermediate",
    "Mi3": "intermediate",
    "Mi4": "intermediate",
    "Mi9": "intermediate",
    "Mi10": "intermediate",
    "Mi11": "intermediate",
    "Mi12": "intermediate",
    "Mi13": "intermediate",
    "Mi14": "intermediate",
    "Mi15": "intermediate",
    "T4a": "output",
    "T4b": "output",
    "T4c": "output",
    "T4d": "output",
    "T1": "output",
    "T2": "output",
    "T2a": "output",
    "T3": "output",
    "T5a": "output",
    "T5b": "output",
    "T5c": "output",
    "T5d": "output",
    "Tm1": "output",
    "Tm2": "output",
    "Tm3": "output",
    "Tm4": "output",
    "Tm5Y": "output",
    "Tm5a": "output",
    "Tm5b": "output",
    "Tm5c": "output",
    "Tm9": "output",
    "Tm16": "output",
    "Tm20": "output",
    "Tm28": "output",
    "Tm30": "output",
    "TmY3": "output",
    "TmY4": "output",
    "TmY5a": "output",
    "TmY9": "output",
    "TmY10": "output",
    "TmY13": "output",
    "TmY14": "output",
    "TmY15": "output",
    "TmY18": "output",
}



"""Script to run ensemble analysis."""

import argparse
import logging
from importlib import resources
from typing import List

from flyvis.utils.compute_cloud_utils import launch_single

logging.basicConfig(
    format="[%(asctime)s] [%(filename)s:%(lineno)d] %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


def run_ensemble_analysis(args: argparse.Namespace, kwargs: List[str]) -> None:
    """
    Launch ensemble analysis job.

    Args:
        args: Command-line arguments.
        kwargs: Additional keyword arguments as a list of strings.
    """
    launch_single(
        args.ensemble_id,
        args.task_name,
        args.nP,
        args.gpu,
        args.q,
        args.ensemble_analysis_script,
        args.dry,
        kwargs,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description=(
            "Run ensemble analysis on the compute cloud. Launches a single job "
            "to analyze "
            "all models in the ensemble."
        ),
        formatter_class=argparse.RawTextHelpFormatter,
        usage=(
            "\nflyvis analysis [-h] [...] --ensemble_id ENSEMBLE_ID "
            "--task_name TASK_NAME "
            "[ensemble_analysis_script_options...]\n"
            "       or\n"
            "%(prog)s [-h] [...] --ensemble_id ENSEMBLE_ID --task_name TASK_NAME "
            "[ensemble_analysis_script_options...]\n"
            "\n"
            "For a full list of options and default arguments, run: "
            "flyvis ensemble-analysis --help"
        ),
    )
    parser.add_argument(
        "--nP", type=int, default=4, help="Number of processors to use (default: 4)."
    )
    parser.add_argument(
        "--gpu", type=str, default="num=1", help="GPU configuration (default: 'num=1')."
    )
    parser.add_argument(
        "--q",
        type=str,
        default="gpu_l4",
        help="Queue to submit the job to.",
    )
    parser.add_argument(
        "--ensemble_id",
        type=int,
        required=True,
        help="ID of the ensemble, e.g., 0045.",
    )
    parser.add_argument(
        "--task_name",
        type=str,
        required=True,
        help="Name given to the task, e.g., 'flow', 'depth', 'lum'.",
    )
    DEFAULT_SCRIPT = str(
        resources.files("flyvis_cli") / "analysis" / "ensemble_analysis.py"
    )
    parser.add_argument(
        "--ensemble_analysis_script",
        type=str,
        default=DEFAULT_SCRIPT,
        help=(f"Script to run for ensemble analysis. Default: {DEFAULT_SCRIPT}"),
    )
    parser.add_argument(
        "--dry",
        action="store_true",
        help="Perform a dry run without actually launching jobs.",
    )

    args, kwargs = parser.parse_known_intermixed_args()
    run_ensemble_analysis(args, kwargs)



"""Connectome compiler and visualizer."""

import json
from contextlib import suppress
from dataclasses import dataclass
from importlib import resources
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
)

import matplotlib.path as mp
import numpy as np
from datamate import ArrayFile, Directory, Namespace, root
from matplotlib import colormaps as cm
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from numpy.typing import NDArray
from pandas import DataFrame
from toolz import groupby, valmap

import flyvis
from flyvis.analysis.visualization import plots, plt_utils
from flyvis.analysis.visualization.figsize_utils import figsize_from_n_items
from flyvis.analysis.visualization.network_fig import WholeNetworkFigure
from flyvis.utils import df_utils, hex_utils, nodes_edges_utils

__all__ = [
    "ConnectomeFromAvgFilters",
    "ConnectomeView",
    "ReceptiveFields",
    "ProjectiveFields",
    "init_connectome",
    "get_avgfilt_connectome",
]


class Connectome(Protocol):
    """Protocol for connectome classes compatible with flyvis.network.Network.

    Note:
        Nodes and edges have additional attributes that require compatibility
        with `Parameter` class implementations. For instance, when a parameter
        for edges is derived from synapse counts, the edges have an `n_syn`
        attribute (ArrayFile or np.ndarray).
    """

    class nodes:
        index: Union[np.ndarray, ArrayFile]
        ...

    class edges:
        source_index: Union[np.ndarray, ArrayFile]
        target_index: Union[np.ndarray, ArrayFile]
        ...


AVAILABLE_CONNECTOMES: Dict[str, Type[Connectome]] = {}


def register_connectome(
    cls: Optional[Type[Connectome]] = None,
) -> Union[Callable[[Type[Connectome]], Type[Connectome]], Type[Connectome]]:
    """
    Register a new connectome class.

    Args:
        cls: The connectome class to register (optional when used as a decorator).

    Returns:
        Registered class or decorator function.

    Example:
        As a standalone function: register_connectome(name, cls)
        ```python
        class CustomConnectome(Connectome):
            ...
        register_connectome("CustomConnectome", CustomConnectome)
        ```

        As a decorator:
        ```python
        @register_connectome("CustomConnectome")
        class CustomConnectome(Connectome):
            ...
        ```
    """

    def decorator(cls: Type[Connectome]) -> Type[Connectome]:
        AVAILABLE_CONNECTOMES[cls.__name__] = cls
        return cls

    return decorator if cls is None else decorator(cls)


# -- `Connectome` --------------------------------------------------------------
@register_connectome
@root(flyvis.root_dir / "connectome")
class ConnectomeFromAvgFilters(Directory):
    """Compiles a connectome graph from average convolutional filters.

    The graph consists of cells (nodes) and synapse sets (edges).

    Args:
        file: The name of a JSON connectome file.
        extent: The array radius, in columns.
        n_syn_fill: The number of synapses to assume in data gaps.

    Attributes:
        unique_cell_types (ArrayFile): Identified cell types.
        input_cell_types (ArrayFile): Input cell types.
        intermediate_cell_types (ArrayFile): Hidden cell types.
        output_cell_types (ArrayFile): Decoded cell types.
        central_cells_index (ArrayFile): Index of central cell in nodes table
            for each cell type in unique_cell_types.
        layout (ArrayFile): Input, hidden, output definitions for visualization.
        nodes (NodeDir): Table with a row for each individual node/cell.
        edges (EdgeDir): Table with a row for each edge.

    Note:
        A connectome can be constructed from a JSON model file following this schema:

        ```json
        {
            "nodes": [{
                "name": string,
                "pattern": (
                    ["stride", [<u_stride:int>, <v_stride:int>]]
                    | ["tile", <stride:int>]
                    | ["single", null]
                )
            }*],
            "edges": [{
                "src": string,
                "tar": string,
                "alpha": int,
                "offsets": [[
                    [<du:int>, <dv:int>],
                    <n_synapses:number>
                    ]*],
                }*]
            }
        }
        ```

        See "data/connectome/fib25-fib19_v2.2.json" for an example.

    Example:
        ```python
        config = Namespace(file='fib25-fib19_v2.2.json', extent=15, n_syn_fill=1)
        connectome = Connectome(config)
        ```
    """

    def __init__(self, file=flyvis.connectome_file.name, extent=15, n_syn_fill=1) -> None:
        # case 0: file is an absolute path
        if Path(file).exists():
            file = Path(file)
        # case 1: file is specified within the package resources
        elif (resources.files("flyvis.connectome") / file).is_file():
            file = resources.files("flyvis.connectome").joinpath(file)
        # case 2: file is specified relative to the root directory
        elif (flyvis.root_dir / "connectome" / file).exists():
            file = flyvis.root_dir / "connectome" / file
        else:
            raise FileNotFoundError(f"Connectome file {file} not found.")

        # Load the connectome spec.
        spec = json.loads(Path(file).read_text())

        # Store unique cell types and layout variables.
        self.unique_cell_types = np.bytes_([n["name"] for n in spec["nodes"]])
        self.input_cell_types = np.bytes_(spec["input_units"])
        self.output_cell_types = np.bytes_(spec["output_units"])
        intermediate_cell_types, _ = nodes_edges_utils.order_node_type_list(
            np.array(
                list(
                    set(self.unique_cell_types)
                    - set(self.input_cell_types)
                    - set(self.output_cell_types)
                )
            ).astype(str)
        )
        self.intermediate_cell_types = np.array(intermediate_cell_types).astype("S")

        layout = []
        layout.extend(
            list(
                zip(
                    self.input_cell_types,
                    [b"retina" for _ in range(len(self.input_cell_types))],
                )
            )
        )
        layout.extend(
            list(
                zip(
                    self.intermediate_cell_types,
                    [b"intermediate" for _ in range(len(self.intermediate_cell_types))],
                )
            )
        )
        layout.extend(
            list(
                zip(
                    self.output_cell_types,
                    [b"output" for _ in range(len(self.output_cell_types))],
                )
            )
        )
        self.layout = np.bytes_(layout)

        # Construct nodes and edges.
        nodes: List[Node] = []
        edges: List[Edge] = []
        add_nodes(nodes, spec["nodes"], extent)
        add_edges(edges, nodes, spec["edges"], n_syn_fill)

        # Define node roles (input, intermediate, output).
        _role = {node: "intermediate" for node in set([n.type for n in nodes])}
        _role.update({node: "input" for node in _role if node in spec["input_units"]})
        _role.update({node: "output" for node in _role if node in spec["output_units"]})

        # Store the graph.
        self.nodes = dict(  # type: ignore
            index=np.int64([n.id for n in nodes]),
            type=np.bytes_([n.type for n in nodes]),
            u=np.int32([n.u for n in nodes]),
            v=np.int32([n.v for n in nodes]),
            role=np.bytes_([_role[n.type] for n in nodes]),
        )

        self.edges = dict(  # type: ignore
            # [Essential fields]
            source_index=np.int64([e.source.id for e in edges]),
            target_index=np.int64([e.target.id for e in edges]),
            sign=np.float32([e.sign for e in edges]),
            n_syn=np.float32([e.n_syn for e in edges]),
            # [Convenience fields]
            source_type=np.bytes_([e.source.type for e in edges]),
            target_type=np.bytes_([e.target.type for e in edges]),
            source_u=np.int32([e.source.u for e in edges]),
            target_u=np.int32([e.target.u for e in edges]),
            source_v=np.int32([e.source.v for e in edges]),
            target_v=np.int32([e.target.v for e in edges]),
            du=np.int32([e.target.u - e.source.u for e in edges]),
            dv=np.int32([e.target.v - e.source.v for e in edges]),
            n_syn_certainty=np.float32([e.n_syn_certainty for e in edges]),
        )

        # Store central indices.
        self.central_cells_index = np.int64(
            np.nonzero((self.nodes.u[:] == 0) & (self.nodes.v[:] == 0))[0]
        )

        # Store layer indices.
        layer_index = {}
        for cell_type in self.unique_cell_types[:]:
            node_indices = np.nonzero(self.nodes["type"][:] == cell_type)[0]
            layer_index[cell_type.decode()] = np.int64(node_indices)
        self.nodes.layer_index = layer_index


# -- Node construction ---------------------------------------------------------


@dataclass
class Node:
    """Represents a node in the connectome graph.

    Attributes:
        id: Index (0..n_nodes).
        type: Cell type name.
        u: Location coordinate #1 (oblique coordinates).
        v: Location coordinate #2 (oblique coordinates).
        u_stride: Stride in u.
        v_stride: Stride in v.
    """

    id: int
    "index (0..n_nodes)"
    type: str
    "cell type name"
    u: int
    "location coordinate #1 (oblique coordinates)"
    v: int
    "location coordinate #2 (oblique coordinates)"
    u_stride: int
    "stride in u"
    v_stride: int
    "stride in v"


class NodeDir(Directory):
    """Stored data of the compiled Connectome describing the nodes."""


def add_nodes(seq: List[Node], node_spec: dict, extent: int) -> None:
    """Add nodes to `seq`, based on `node_spec`.

    Args:
        seq: List to append nodes to.
        node_spec: Dictionary specifying node properties.
        extent: The array radius, in columns.
    """
    for n in node_spec:
        typ, (pattern, args) = n["name"], n["pattern"]
        if pattern == "stride":
            add_strided_nodes(seq, typ, extent, args)
        elif pattern == "tile":
            add_tiled_nodes(seq, typ, extent, args)
        elif pattern == "single":
            add_single_node(seq, typ, extent)


def add_strided_nodes(
    seq: List[Node], typ: str, extent: int, strides: Tuple[int, int]
) -> None:
    """Add to `seq` a population of neurons arranged in a hexagonal grid.

    Args:
        seq: List to append nodes to.
        typ: Cell type name.
        extent: The array radius, in columns.
        strides: Tuple of (u_stride, v_stride).
    """
    n = extent
    u_stride, v_stride = strides
    for u in range(-n, n + 1):
        for v in range(max(-n, -n - u), min(n, n - u) + 1):
            if u % u_stride == 0 and v % v_stride == 0:
                seq.append(Node(len(seq), typ, u, v, u_stride, v_stride))


def add_tiled_nodes(seq: List[Node], typ: str, extent: int, n: int) -> None:
    """Add to `seq` a population of neurons with strides `(n, n)`.

    Args:
        seq: List to append nodes to.
        typ: Cell type name.
        extent: The array radius, in columns.
        n: Stride value for both u and v.
    """
    add_strided_nodes(seq, typ, extent, (n, n))


def add_single_node(seq: List[Node], typ: str, extent: int) -> None:
    """Add to `seq` a single-neuron population.

    Args:
        seq: List to append nodes to.
        typ: Cell type name.
        extent: The array radius, in columns.
    """
    add_strided_nodes(seq, typ, 0, (1, 1))


# -- Edge construction ---------------------------------------------------------


@dataclass
class Edge:
    """Represents an edge in the connectome graph.

    Attributes:
        id: Index (0..n_edges).
        source: Presynaptic cell.
        target: Postsynaptic cell.
        sign: +1 (excitatory) or -1 (inhibitory).
        n_syn: Synapse count.
        type: Synapse type.
        n_syn_certainty: Certainty of synapse count.
    """

    id: int
    "index (0..n_edges)"
    source: Node
    "presynaptic cell"
    target: Node
    "postsynaptic cell"
    sign: int
    "+1 (excitatory) or -1 (inhibitory)"
    n_syn: float
    "synapse count"
    n_syn_certainty: float
    "certainty of synapse count"


class EdgeDir(Directory):
    """Stored data of the compiled Connectome describing the edges."""


def add_edges(
    seq: List[Edge], nodes: List[Node], edge_spec: dict, n_syn_fill: float
) -> None:
    """Add edges to `seq` based on `edge_spec`.

    Args:
        seq: List to append edges to.
        nodes: List of all nodes in the connectome.
        edge_spec: Dictionary specifying edge properties.
        n_syn_fill: Number of synapses to assume in data gaps.
    """
    node_index = {
        **groupby(lambda n: n.type, nodes),
        **groupby(lambda n: (n.type, n.u, n.v), nodes),
    }
    for e in edge_spec:
        offsets = (
            fill_hull(e["offsets"], n_syn_fill)
            if n_syn_fill > 0 and len(e["offsets"]) >= 3
            else e["offsets"]
        )
        add_conv_edges(
            seq,
            node_index,
            e["src"],
            e["tar"],
            e["alpha"],
            offsets,
            e["lambda_mult"],
        )


def fill_hull(offsets: List[list], n_syn_fill: float) -> List[list]:
    """Fill in the convex hull of the reported edges.

    Args:
        offsets: List of edge offsets.
        n_syn_fill: Number of synapses to assume in data gaps.

    Returns:
        List of filled offsets.
    """
    # to save time at library import
    import scipy.spatial as ss

    # Collect the points (column offsets, (du, dv)) reported as edges.
    known_pts = np.array([offset[0] for offset in offsets])
    known_pts_as_set = set(map(tuple, known_pts))

    # Compute the convex hull of the reported edges.
    hull = ss.ConvexHull((1 + 1e-6) * known_pts, False, "QJ")
    hull_vertices = known_pts[hull.vertices]

    # Find the points within the convex hull.
    grid = np.concatenate(
        np.dstack(
            np.mgrid[
                known_pts[:, 0].min() : known_pts[:, 0].max() + 1,
                known_pts[:, 1].min() : known_pts[:, 1].max() + 1,
            ]
        )
    )
    contained_pts = grid[mp.Path(hull_vertices).contains_points(grid)]

    # Add unknown points within the convex hull as offsets.
    return offsets + [
        [[u, v], n_syn_fill] for u, v in contained_pts if (u, v) not in known_pts_as_set
    ]


def add_conv_edges(
    seq: List[Edge],
    node_index: dict,
    source_typ: str,
    target_typ: str,
    sign: int,
    offsets: List[list],
    n_syn_certainty: float,
) -> None:
    """Construct a connection set with convolutional weight symmetry.

    Args:
        seq: List to append edges to.
        node_index: Dictionary mapping node types to nodes.
        source_typ: Source cell type.
        target_typ: Target cell type.
        sign: +1 (excitatory) or -1 (inhibitory).
        offsets: List of edge offsets.
        n_syn_certainty: Certainty of synapse count.
    """
    for (du, dv), n_syn in offsets:
        for src in node_index.get(source_typ, []):
            u_tgt = src.u + du
            v_tgt = src.v + dv
            with suppress(KeyError):
                tgt = node_index[target_typ, u_tgt, v_tgt][0]
                seq.append(Edge(len(seq), src, tgt, sign, n_syn, n_syn_certainty))


# -- ConnectomeView ------------------------------------------------------------


class ConnectomeView:
    """Visualization of the connectome data.

    Args:
        connectome: Directory of the connectome.
        groups: Regular expressions to sort the nodes by.

    Attributes:
        dir (ConnectomeFromAvgFilters): Connectome directory.
        edges (Directory): Edge table.
        nodes (Directory): Node table.
        cell_types_unsorted (List[str]): Unsorted list of cell types.
        cell_types_sorted (List[str]): Sorted list of cell types.
        cell_types_sort_index (List[int]): Indices for sorting cell types.
        layout (Dict[str, str]): Layout information for cell types.
        node_indexer (NodeIndexer): Indexer for nodes.
    """

    def __init__(
        self,
        connectome: ConnectomeFromAvgFilters,
        groups: List[str] = [
            r"R\d",
            r"L\d",
            r"Lawf\d",
            r"A",
            r"C\d",
            r"CT\d.*",
            r"Mi\d{1,2}",
            r"T\d{1,2}.*",
            r"Tm.*\d{1,2}.*",
        ],
    ) -> None:
        self.dir = connectome

        assert "nodes" in self.dir and "edges" in self.dir

        self.edges = self.dir.edges
        self.nodes = self.dir.nodes

        self.cell_types_unsorted = self.dir.unique_cell_types[:].astype(str)

        (
            self.cell_types_sorted,
            self.cell_types_sort_index,
        ) = nodes_edges_utils.order_node_type_list(
            self.dir.unique_cell_types[:].astype(str), groups
        )

        self.layout = dict(self.dir.layout[:].astype(str))
        self.node_indexer = nodes_edges_utils.NodeIndexer(self.dir)

    def connectivity_matrix(
        self,
        mode: str = "n_syn",
        only_sign: Optional[str] = None,
        cell_types: Optional[List[str]] = None,
        no_symlog: Optional[bool] = False,
        min_number: Optional[float] = None,
        cmap: Optional[Colormap] = None,
        size_scale: Optional[float] = None,
        title: Optional[str] = None,
        cbar_label: Optional[str] = None,
        **kwargs,
    ) -> Figure:
        """Plot the connectivity matrix as counts or weights.

        Args:
            mode: 'n_syn' for number of input synapses, 'count' for number of neurons.
            only_sign: '+' for excitatory projections, '-' for inhibitory projections.
            cell_types: Subset of nodes to display.
            no_symlog: Disable symmetric log scale.
            min_number: Minimum value to display.
            cmap: Custom colormap.
            size_scale: Size of the scattered squares.
            title: Custom title for the plot.
            cbar_label: Custom colorbar label.
            **kwargs: Additional arguments passed to the heatmap plot function.

        Returns:
            Figure: Matplotlib figure object.
        """
        _kwargs = dict(
            n_syn=dict(
                symlog=1e-5,
                grid=True,
                cmap=cmap or cm.get_cmap("seismic"),
                title=title or "Connectivity between identified cell types",
                cbar_label=cbar_label or r"$\pm\sum_{pre} N_\mathrm{syn.}^{pre, post}$",
                size_scale=size_scale or 0.05,
            ),
            count=dict(
                grid=True,
                cmap=cmap or cm.get_cmap("seismic"),
                midpoint=0,
                title=title or "Number of Input Neurons",
                cbar_label=cbar_label or r"$\sum_{pre} 1$",
                size_scale=size_scale or 0.05,
            ),
        )

        kwargs.update(_kwargs[mode])
        if no_symlog:
            kwargs.update(symlog=None)
            kwargs.update(midpoint=0)

        edges = self.edges.to_df()

        # to take projections onto central nodes (home columns) into account
        edges = edges[(edges.target_u == 0) & (edges.target_v == 0)]

        # filter edges to allow providing a subset of cell types
        cell_types = cell_types or self.cell_types_sorted
        edges = df_utils.filter_by_column_values(
            df_utils.filter_by_column_values(
                edges, column="source_type", values=cell_types
            ),
            column="target_type",
            values=cell_types,
        )
        weights = self._weights()[edges.index]

        # lookup table for key -> (i, j)
        type_index = {node_typ: i for i, node_typ in enumerate(cell_types)}
        matrix = np.zeros([len(type_index), len(type_index)])

        for srctyp, tgttyp, weight in zip(
            edges.source_type.values, edges.target_type.values, weights
        ):
            if mode == "count":
                # to simply count the number of projections
                matrix[type_index[srctyp], type_index[tgttyp]] += 1
            elif mode in ["weight", "n_syn"]:
                # to sum the synapse counts
                matrix[type_index[srctyp], type_index[tgttyp]] += weight
            else:
                raise ValueError

        # to filter out all connections weaker than min_number
        if min_number is not None:
            matrix[np.abs(matrix) <= min_number] = np.nan

        # to display either only excitatory or inhibitory connections
        if only_sign == "+":
            matrix[matrix < 0] = 0
            kwargs.update(symlog=None, midpoint=0)
        elif only_sign == "-":
            matrix[matrix > 0] = 0
            kwargs.update(symlog=None, midpoint=0)
        elif only_sign is None:
            pass
        else:
            raise ValueError

        return plots.heatmap(matrix, cell_types, **kwargs)

    def _weights(self) -> NDArray:
        """Calculate weights for edges.

        Returns:
            NDArray: Array of edge weights.
        """
        return self.edges.sign[:] * self.edges.n_syn[:]

    def network_layout(
        self,
        max_extent: int = 5,
        **kwargs,
    ) -> Figure:
        """Plot retinotopic hexagonal lattice columnar organization of the network.

        Args:
            max_extent: Integer column radius to visualize.
            **kwargs: Additional arguments passed to hex_layout_all.

        Returns:
            Figure: Matplotlib figure object.
        """
        backbone = WholeNetworkFigure(self.dir)
        backbone.init_figure(figsize=[7, 3])
        return self.hex_layout_all(
            max_extent=max_extent, fig=backbone.fig, axes=backbone.axes, **kwargs
        )

    def hex_layout(
        self,
        cell_type: str,
        max_extent: int = 5,
        edgecolor: str = "none",
        edgewidth: float = 0.5,
        alpha: float = 1,
        fill: bool = False,
        cmap: Optional[Colormap] = None,
        fig: Optional[Figure] = None,
        ax: Optional[Axes] = None,
        **kwargs,
    ) -> Figure:
        """Plot retinotopic hexagonal lattice organization of a cell type.

        Args:
            cell_type: Type of cell to plot.
            max_extent: Maximum extent of the layout.
            edgecolor: Color of the hexagon edges.
            edgewidth: Width of the hexagon edges.
            alpha: Transparency of the hexagons.
            fill: Whether to fill the hexagons.
            cmap: Custom colormap.
            fig: Existing figure to plot on.
            ax: Existing axis to plot on.
            **kwargs: Additional arguments passed to hex_scatter.

        Returns:
            Figure: Matplotlib figure object.
        """
        nodes = self.nodes.to_df()
        node_condition = nodes.type == cell_type
        u, v = nodes.u[node_condition], nodes.v[node_condition]
        max_extent = hex_utils.get_extent(u, v) if max_extent is None else max_extent
        extent_condition = (
            (-max_extent <= u)
            & (u <= max_extent)
            & (-max_extent <= v)
            & (v <= max_extent)
            & (-max_extent <= u + v)
            & (u + v <= max_extent)
        )
        u, v = u[extent_condition].values, v[extent_condition].values

        label = cell_type
        if ax is not None:
            # prevent labeling twice
            label = cell_type if cell_type not in [t.get_text() for t in ax.texts] else ""

        fig, ax, _ = plots.hex_scatter(
            u,
            v,
            values=1,
            label=label,
            fig=fig,
            ax=ax,
            edgecolor=edgecolor,
            edgewidth=edgewidth,
            alpha=alpha,
            fill=fill,
            cmap=cmap or plt_utils.get_alpha_colormap("#2f3541", 1),
            cbar=False,
            **kwargs,
        )
        return fig

    def hex_layout_all(
        self,
        cell_types: Optional[List[str]] = None,
        max_extent: int = 5,
        edgecolor: str = "none",
        alpha: float = 1,
        fill: bool = False,
        cmap: Optional[Colormap] = None,
        fig: Optional[Figure] = None,
        axes: Optional[List[Axes]] = None,
        **kwargs,
    ) -> Figure:
        """Plot retinotopic hexagonal lattice organization of all cell types.

        Args:
            cell_types: List of cell types to plot.
            max_extent: Maximum extent of the layout.
            edgecolor: Color of the hexagon edges.
            alpha: Transparency of the hexagons.
            fill: Whether to fill the hexagons.
            cmap: Custom colormap.
            fig: Existing figure to plot on.
            axes: List of existing axes to plot on.
            **kwargs: Additional arguments passed to hex_layout.

        Returns:
            Figure: Matplotlib figure object.
        """
        cell_types = self.cell_types_sorted if cell_types is None else cell_types
        if fig is None or axes is None:
            fig, axes, (gw, gh) = plt_utils.get_axis_grid(self.cell_types_sorted)

        for i, cell_type in enumerate(cell_types):
            self.hex_layout(
                cell_type,
                edgecolor=edgecolor,
                edgewidth=0.1,
                alpha=alpha,
                fill=fill,
                max_extent=max_extent,
                cmap=cmap or plt_utils.get_alpha_colormap("#2f3541", 1),
                fig=fig,
                ax=axes[i],
                **kwargs,
            )
        return fig

    def get_uv(self, cell_type: str) -> Tuple[NDArray, NDArray]:
        """Get hex-coordinates of a particular cell type.

        Args:
            cell_type: Type of cell to get coordinates for.

        Returns:
            Tuple[NDArray, NDArray]: Arrays of u and v coordinates.
        """
        nodes = self.nodes.to_df()
        nodes = nodes[nodes.type == cell_type]
        u, v = nodes[["u", "v"]].values.T
        return u, v

    def sources_list(self, cell_type: str) -> NDArray:
        """Get presynaptic cell types.

        Args:
            cell_type: Type of cell to get sources for.

        Returns:
            NDArray: Array of presynaptic cell types.
        """
        edges = self.edges.to_df()
        return np.unique(edges[edges.target_type == cell_type].source_type.values)

    def targets_list(self, cell_type: str) -> NDArray:
        """Get postsynaptic cell types.

        Args:
            cell_type: Type of cell to get targets for.

        Returns:
            NDArray: Array of postsynaptic cell types.
        """
        edges = self.edges.to_df()
        return np.unique(edges[edges.source_type == cell_type].target_type.values)

    def receptive_field(
        self,
        source: str = "Mi9",
        target: str = "T4a",
        rfs: Optional["ReceptiveFields"] = None,
        max_extent: Optional[int] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        title: str = "{source} :→ {target}",
        **kwargs,
    ) -> Figure:
        """Plot the receptive field of a target cell type from a source cell type.

        Args:
            source: Source cell type.
            target: Target cell type.
            rfs: ReceptiveFields object. If None, it will be created.
            max_extent: Maximum extent of the receptive field.
            vmin: Minimum value for colormap.
            vmax: Maximum value for colormap.
            title: Title format string for the plot.
            **kwargs: Additional arguments passed to plots.kernel.

        Returns:
            Matplotlib Figure object.
        """
        if rfs is None:
            rfs = ReceptiveFields(target, self.edges.to_df())
            max_extent = max_extent or rfs.max_extent

        weights = self._weights()

        # to derive color range values taking all inputs into account
        vmin = min(
            0,
            min(weights[rfs[source].index].min() for source in rfs.source_types),
        )
        vmax = max(
            0,
            max(weights[rfs[source].index].max() for source in rfs.source_types),
        )

        weights = weights[rfs[source].index]
        label = ""

        # requires to look from the target cell, ie mirror the coordinates
        du_inv, dv_inv = -rfs[source].du.values, -rfs[source].dv.values
        fig, ax, (label_text, scalarmapper) = plots.kernel(
            du_inv,
            dv_inv,
            weights,
            label=label,
            max_extent=max_extent,
            fill=True,
            vmin=vmin,
            vmax=vmax,
            title=title.format(**locals()),
            **kwargs,
        )
        return fig

    def receptive_fields_grid(
        self,
        target: str,
        sources: Optional[Iterable[str]] = None,
        sort_alphabetically: bool = True,
        ax_titles: str = "{source} :→ {target}",
        figsize: List[int] = [20, 20],
        max_extent: Optional[int] = None,
        fig: Optional[Figure] = None,
        axes: Optional[List[Axes]] = None,
        ignore_sign_error: bool = False,
        max_figure_height_cm: float = 22,
        panel_height_cm: float = 3,
        max_figure_width_cm: float = 18,
        panel_width_cm: float = 3.6,
        **kwargs,
    ) -> Figure:
        """Plot receptive fields of a target cell type in a grid layout.

        Args:
            target: Target cell type.
            sources: Iterable of source cell types. If None, all sources are used.
            sort_alphabetically: Whether to sort source types alphabetically.
            ax_titles: Title format string for each subplot.
            figsize: Figure size in inches.
            max_extent: Maximum extent of the receptive fields.
            fig: Existing figure to plot on.
            axes: List of existing axes to plot on.
            ignore_sign_error: Whether to ignore sign errors in plotting.
            max_figure_height_cm: Maximum figure height in cm.
            panel_height_cm: Height of each panel in cm.
            max_figure_width_cm: Maximum figure width in cm.
            panel_width_cm: Width of each panel in cm.
            **kwargs: Additional arguments passed to receptive_field.

        Returns:
            Matplotlib Figure object.
        """

        rfs = ReceptiveFields(target, self.edges.to_df())
        max_extent = max_extent or rfs.max_extent
        weights = self._weights()

        # to sort in descending order by sum of inputs
        sorted_sum_of_inputs = dict(
            sorted(
                valmap(lambda v: weights[v.index].sum(), rfs).items(),
                key=lambda item: item[1],
                reverse=True,
            )
        )
        # to sort alphabetically in case sources is specified
        if sort_alphabetically:
            sources, _ = nodes_edges_utils.order_node_type_list(sources)
        sources = sources or list(sorted_sum_of_inputs.keys())

        # to derive color range values taking all inputs into account
        vmin = min(0, min(weights[rfs[source].index].min() for source in sources))
        vmax = max(0, max(weights[rfs[source].index].max() for source in sources))

        if fig is None or axes is None:
            figsize = figsize_from_n_items(
                len(rfs.source_types),
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=panel_height_cm,
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(
                unmask_n=len(rfs.source_types), hspace=0.0, wspace=0
            )

        cbar = kwargs.get("cbar", False)
        for i, src in enumerate(sources):
            if i == 0 and cbar:
                cbar = True
                kwargs.update(cbar=cbar)
            else:
                cbar = False
                kwargs.update(cbar=cbar)
            try:
                self.receptive_field(
                    target=target,
                    source=src,
                    fig=fig,
                    ax=axes[i],
                    title=ax_titles,
                    vmin=vmin,
                    vmax=vmax,
                    rfs=rfs,
                    max_extent=max_extent,
                    annotate=False,
                    annotate_coords=False,
                    title_y=0.9,
                    **kwargs,
                )
            except plots.SignError as e:
                if ignore_sign_error:
                    pass
                else:
                    raise e
        return fig

    def projective_field(
        self,
        source: str = "Mi9",
        target: str = "T4a",
        title: str = "{source} →: {target}",
        prfs: Optional["ProjectiveFields"] = None,
        max_extent: Optional[int] = None,
        vmin: Optional[float] = None,
        vmax: Optional[float] = None,
        **kwargs,
    ) -> Optional[Figure]:
        """Plot the projective field from a source cell type to a target cell type.

        Args:
            source: Source cell type.
            target: Target cell type.
            title: Title format string for the plot.
            prfs: ProjectiveFields object. If None, it will be created.
            max_extent: Maximum extent of the projective field.
            vmin: Minimum value for colormap.
            vmax: Maximum value for colormap.
            **kwargs: Additional arguments passed to plots.kernel.

        Returns:
            Matplotlib Figure object or None if max_extent is None.
        """
        if prfs is None:
            prfs = ProjectiveFields(source, self.edges.to_df())
            max_extent = max_extent or prfs.max_extent
        if max_extent is None:
            return None
        weights = self._weights()

        # to derive color range values taking all inputs into account
        vmin = min(
            0,
            min(weights[prfs[target].index].min() for target in prfs.target_types),
        )

        vmax = max(
            0,
            max(weights[prfs[target].index].max() for target in prfs.target_types),
        )

        weights = weights[prfs[target].index]
        label = ""
        du, dv = prfs[target].du.values, prfs[target].dv.values
        fig, ax, (label_text, scalarmapper) = plots.kernel(
            du,
            dv,
            weights,
            label=label,
            fill=True,
            max_extent=max_extent,
            vmin=vmin,
            vmax=vmax,
            title=title.format(**locals()),
            **kwargs,
        )
        return fig

    def projective_fields_grid(
        self,
        source: str,
        targets: Optional[Iterable[str]] = None,
        fig: Optional[Figure] = None,
        axes: Optional[List[Axes]] = None,
        figsize: List[int] = [20, 20],
        ax_titles: str = "{source} →: {target}",
        max_figure_height_cm: float = 22,
        panel_height_cm: float = 3,
        max_figure_width_cm: float = 18,
        panel_width_cm: float = 3.6,
        max_extent: Optional[int] = None,
        sort_alphabetically: bool = False,
        ignore_sign_error: bool = False,
        **kwargs,
    ) -> Figure:
        """Plot projective fields of a source cell type in a grid layout.

        Args:
            source: Source cell type.
            targets: Iterable of target cell types. If None, all targets are used.
            fig: Existing figure to plot on.
            axes: List of existing axes to plot on.
            figsize: Figure size in inches.
            ax_titles: Title format string for each subplot.
            max_figure_height_cm: Maximum figure height in cm.
            panel_height_cm: Height of each panel in cm.
            max_figure_width_cm: Maximum figure width in cm.
            panel_width_cm: Width of each panel in cm.
            max_extent: Maximum extent of the projective fields.
            sort_alphabetically: Whether to sort target types alphabetically.
            ignore_sign_error: Whether to ignore sign errors in plotting.
            **kwargs: Additional arguments passed to projective_field.

        Returns:
            Matplotlib Figure object.
        """
        prfs = ProjectiveFields(source, self.edges.to_df())
        max_extent = max_extent or prfs.max_extent
        weights = self._weights()
        sorted_sum_of_outputs = dict(
            sorted(
                valmap(lambda v: weights[v.index].sum(), prfs).items(),
                key=lambda item: item[1],
                reverse=True,
            )
        )

        # to sort alphabetically in case sources is specified
        if sort_alphabetically:
            targets, _ = nodes_edges_utils.order_node_type_list(targets)

        targets = targets or list(sorted_sum_of_outputs.keys())

        vmin = min(0, min(weights[prfs[target].index].min() for target in targets))
        vmax = max(0, max(weights[prfs[target].index].max() for target in targets))

        if fig is None or axes is None:
            figsize = figsize_from_n_items(
                len(prfs.target_types),
                max_figure_height_cm=max_figure_height_cm,
                panel_height_cm=panel_height_cm,
                max_figure_width_cm=max_figure_width_cm,
                panel_width_cm=panel_width_cm,
            )
            fig, axes = figsize.axis_grid(
                unmask_n=len(prfs.target_types), hspace=0.0, wspace=0
            )

        cbar = kwargs.get("cbar", False)
        for i, target in enumerate(targets):
            if i == 0 and cbar:
                cbar = True
                kwargs.update(cbar=cbar)
            else:
                cbar = False
                kwargs.update(cbar=cbar)
            try:
                self.projective_field(
                    source=source,
                    target=target,
                    fig=fig,
                    ax=axes[i],
                    title=ax_titles,
                    prfs=prfs,
                    max_extent=max_extent,
                    vmin=vmin,
                    vmax=vmax,
                    annotate_coords=False,
                    annotate=False,
                    title_y=0.9,
                    **kwargs,
                )
            except plots.SignError as e:
                if ignore_sign_error:
                    pass
                else:
                    raise e
        return fig

    def receptive_fields_df(self, target_type: str) -> "ReceptiveFields":
        """Get receptive fields for a target cell type.

        Args:
            target_type: Target cell type.

        Returns:
            ReceptiveFields object.
        """
        return ReceptiveFields(target_type, self.edges.to_df())

    def projective_fields_df(self, source_type: str) -> "ProjectiveFields":
        """Get projective fields for a source cell type.

        Args:
            source_type: Source cell type.

        Returns:
            ProjectiveFields object.
        """
        return ProjectiveFields(source_type, self.edges.to_df())

    def receptive_fields_sum(self, target_type: str) -> Dict[str, int]:
        """Get sum of synapses for each source type in the receptive field.

        Args:
            target_type: Target cell type.

        Returns:
            Dictionary mapping source types to synapse counts.
        """
        return ReceptiveFields(target_type, self.edges.to_df()).sum()

    def projective_fields_sum(self, source_type: str) -> Dict[str, int]:
        """Get sum of synapses for each target type in the projective field.

        Args:
            source_type: Source cell type.

        Returns:
            Dictionary mapping target types to synapse counts.
        """


class ReceptiveFields(Namespace):
    """Dictionary of receptive field dataframes for a specific cell type.

    Args:
        target_type: Target cell type.
        edges: All edges of a Connectome.

    Attributes:
        target_type: The target cell type.
        source_types: List of source cell types.
        _extents: List of extents for each source type.

    Example:
        ```python
        rf = ReceptiveFields("T4a", edges_dataframe)
        ```
    """

    def __init__(self, target_type: str, edges: DataFrame, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, "_extents", [])
        _receptive_fields_edge_dfs(self, target_type, edges)

    @property
    def extents(self) -> Dict[str, int]:
        """Dictionary of extents for each source type."""
        return dict(zip(self.source_types, self._extents))

    @property
    def max_extent(self) -> Optional[int]:
        """Maximum extent across all source types."""
        return max(self._extents) if self._extents else None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.target_type})"

    def sum(self) -> Dict[str, float]:
        """Sum of synapses for each source type."""
        return {key: self[key].n_syn.sum() for key in self}


class ProjectiveFields(Namespace):
    """Dictionary of projective field dataframes for a specific cell type.

    Args:
        source_type: Source cell type.
        edges: All edges of a Connectome.

    Attributes:
        source_type: The source cell type.
        target_types: List of target cell types.
        _extents: List of extents for each target type.

    Example:
        ```python
        pf = ProjectiveFields("Mi9", edges_dataframe)
        ```
    """

    def __init__(self, source_type: str, edges: DataFrame, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, "_extents", [])
        _projective_fields_edge_dfs(self, source_type, edges)

    @property
    def extents(self) -> Dict[str, int]:
        """Dictionary of extents for each target type."""
        return dict(zip(self.target_types, self._extents))

    @property
    def max_extent(self) -> Optional[int]:
        """Maximum extent across all target types."""
        return max(self._extents) if self._extents else None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.source_type})"

    def sum(self) -> Dict[str, float]:
        """Sum of synapses for each target type."""
        return {key: self[key].n_syn.sum() for key in self}


def _receptive_fields_edge_dfs(
    cls: ReceptiveFields, target_type: str, edges: DataFrame
) -> ReceptiveFields:
    """Populate ReceptiveFields with edge dataframes.

    Args:
        cls: ReceptiveFields instance to populate.
        target_type: Target cell type.
        edges: DataFrame containing all edges.

    Returns:
        Populated ReceptiveFields instance.
    """
    edges = edges[edges.target_type == target_type]
    source_types = edges.source_type.unique()

    object.__setattr__(cls, "target_type", target_type)
    object.__setattr__(cls, "source_types", source_types)

    for source_type in source_types:
        _edges = edges[edges.source_type == source_type]

        most_central_edge = _edges.iloc[
            np.argmin(np.abs(_edges.target_u) + np.abs(_edges.target_v))
        ]
        target_u_min = most_central_edge.target_u
        target_v_min = most_central_edge.target_v

        cls[source_type] = _edges[
            (_edges.target_u == target_u_min) & (_edges.target_v == target_v_min)
        ]
        cls._extents.append(
            hex_utils.get_extent(cls[source_type].du, cls[source_type].dv)
        )
    return cls


def _projective_fields_edge_dfs(
    cls: ProjectiveFields, source_type: str, edges: DataFrame
) -> ProjectiveFields:
    """Populate ProjectiveFields with edge dataframes.

    Args:
        cls: ProjectiveFields instance to populate.
        source_type: Source cell type.
        edges: DataFrame containing all edges.

    Returns:
        Populated ProjectiveFields instance.
    """
    edges = edges[edges.source_type == source_type]
    target_types = edges.target_type.unique()

    object.__setattr__(cls, "source_type", source_type)
    object.__setattr__(cls, "target_types", target_types)

    for target_type in target_types:
        _edges = edges[edges.target_type == target_type]
        most_central_edge = _edges.iloc[
            np.argmin(np.abs(_edges.source_u) + np.abs(_edges.source_v))
        ]
        source_u_min = most_central_edge.source_u
        source_v_min = most_central_edge.source_v
        cls[target_type] = _edges[
            (_edges.source_u == source_u_min) & (_edges.source_v == source_v_min)
        ]
        cls._extents.append(
            hex_utils.get_extent(cls[target_type].du, cls[target_type].dv)
        )

    return cls


def get_avgfilt_connectome(config: dict) -> ConnectomeView:
    """Create a ConnectomeView instance from a config for ConnectomeFromAvgFilters.

    Args:
        config: Containing ConnectomeFromAvgFilters configuration.

    Returns:
        ConnectomeView instance.
    """
    return ConnectomeView(ConnectomeFromAvgFilters(**config))


def is_connectome_protocol(obj: Any) -> bool:
    """
    Check if an object implements the Connectome(Protocol).

    Args:
        obj: The object to check.

    Returns:
        bool: True if the object implements the Connectome(Protocol), False otherwise.

    Note:
        The Connectome(Protocol) requires the following attributes:
        - nodes.index: Union[np.ndarray, ArrayFile]
        - edges.source_index: Union[np.ndarray, ArrayFile]
        - edges.target_index: Union[np.ndarray, ArrayFile]
    """
    if not hasattr(obj, 'nodes'):
        return False, "Missing 'nodes' attribute"
    if not hasattr(obj.nodes, 'index'):
        return False, "Missing 'nodes.index' attribute"
    if not isinstance(obj.nodes.index, (np.ndarray, ArrayFile)):
        return False, "'nodes.index' is not of type np.ndarray or ArrayFile"
    if not hasattr(obj, 'edges'):
        return False, "Missing 'edges' attribute"
    if not hasattr(obj.edges, 'source_index'):
        return False, "Missing 'edges.source_index' attribute"
    if not isinstance(obj.edges.source_index, (np.ndarray, ArrayFile)):
        return False, "'edges.source_index' is not of type np.ndarray or ArrayFile"
    if not hasattr(obj.edges, 'target_index'):
        return False, "Missing 'edges.target_index' attribute"
    if not isinstance(obj.edges.target_index, (np.ndarray, ArrayFile)):
        return False, "'edges.target_index' is not of type np.ndarray or ArrayFile"
    return True, ""


def init_connectome(**kwargs) -> Connectome:
    """Initialize a Connectome instance from a config dictionary.

    Args:
        config: A dictionary containing the connectome configuration.

    Returns:
        An instance of a class implementing the Connectome(Protocol).

    Raises:
        KeyError: If the specified connectome type is not available.

    Example:
        ```python
        config = {
            "type": "ConnectomeFromAvgFilters",
            **config
        }
        connectome = init_connectome(**config)
        ```
    """
    connectome_class = AVAILABLE_CONNECTOMES[kwargs.pop("type")]

    connectome = connectome_class(**kwargs)
    is_valid, error_msg = is_connectome_protocol(connectome)
    assert is_valid, (
        f"Connectome class {connectome} does "
        f"not implement the Connectome(Protocol): {error_msg}"
    )
    return connectome



"""'Transduction' of cartesian pixels to hexals on a regular hexagonal lattice."""

from itertools import product
from typing import Iterator, List, Literal, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn.functional as F
import torchvision.transforms.functional as ttf
from torch import nn

import flyvis
from flyvis.analysis.visualization.plt_utils import init_plot, rm_spines

from .utils import (
    hex_center_coordinates,
    is_inside_hex,
    median,
    render_bars_cartesian,
    render_gratings_cartesian,
)

__all__ = ["BoxEye", "HexEye"]

# ----- BoxEye -----------------------------------------------------------------


class BoxEye:
    """BoxFilter to produce an array of hexals matching the photoreceptor array.

    Args:
        extent: Radius, in number of receptors, of the hexagonal array.
        kernel_size: Photon collection radius, in pixels.

    Attributes:
        extent (int): Radius, in number of receptors, of the hexagonal array.
        kernel_size (int): Photon collection radius, in pixels.
        receptor_centers (torch.Tensor): Tensor of shape (hexals, 2) containing the y, x
            coordinates of the hexal centers.
        hexals (int): Number of hexals in the array.
        min_frame_size (torch.Tensor): Minimum frame size to contain the hexal array.
        pad (Tuple[int, int, int, int]): Padding to apply to the frame before convolution.
        conv (nn.Conv2d): Convolutional box filter to apply to the frame.
    """

    def __init__(self, extent: int = 15, kernel_size: int = 13):
        self.extent = extent
        self.kernel_size = kernel_size
        self.receptor_centers = torch.tensor(
            [*self._receptor_centers()], dtype=torch.long
        )
        self.hexals = len(self.receptor_centers)
        # The rest of kernel_size distance from outer centers to the border
        # is taken care of by the padding of the convolution object.
        self.min_frame_size = (
            self.receptor_centers.max(dim=0).values
            - self.receptor_centers.min(dim=0).values
            + 1
        )
        self._set_filter()

        pad = (self.kernel_size - 1) / 2
        self.pad = (
            int(np.ceil(pad)),
            int(np.floor(pad)),
            int(np.ceil(pad)),
            int(np.floor(pad)),
        )

    def _receptor_centers(self) -> Iterator[Tuple[float, float]]:
        """Generate receptor center coordinates.

        Returns:
            Iterator[Tuple[float, float]]: Yields y, x coordinates of receptor centers.
        """
        n = self.extent
        d = self.kernel_size
        for u in range(-n, n + 1):
            v_min = max(-n, -n - u)
            v_max = min(n, n - u)
            for v in range(v_min, v_max + 1):
                # y = -d * v
                # x = 2 / np.sqrt(3) * d * (u + v/2)
                y = d * (
                    u + v / 2
                )  # - d * v # either must be negative or origin must be upper
                x = d * v  # 2 / np.sqrt(3) * d * (u + v / 2)
                yield y, x
                # xs.append()
                # ys.append()

    def _set_filter(self) -> None:
        """Set up the convolutional filter for the box kernel."""
        self.conv = nn.Conv2d(1, 1, kernel_size=self.kernel_size, stride=1, padding=0)
        self.conv.weight.data /= self.conv.weight.data
        self.conv.bias.data.fill_(0)  # if not self.requires_grad else None
        self.conv.weight.requires_grad = False  # self.requires_grad
        self.conv.bias.requires_grad = False  # self.requires_grad

    def __call__(
        self,
        sequence: torch.Tensor,
        ftype: Literal["mean", "sum", "median"] = "mean",
        hex_sample: bool = True,
    ) -> torch.Tensor:
        """Apply a box kernel to all frames in a sequence.

        Args:
            sequence: Cartesian movie sequences of shape (samples, frames, height, width).
            ftype: Filter type.
            hex_sample: If False, returns filtered cartesian sequences.

        Returns:
            torch.Tensor: Shape (samples, frames, 1, hexals) if hex_sample is True,
                otherwise (samples, frames, height, width).
        """
        samples, frames, height, width = sequence.shape

        if not isinstance(sequence, torch.Tensor):
            # auto-moving to GPU in case default tensor is cuda but passed
            # sequence is not, for convenience
            sequence = torch.tensor(sequence, dtype=torch.float32, device=flyvis.device)

        if (self.min_frame_size > torch.tensor([height, width])).any():
            # to rescale to the minimum frame size
            sequence = ttf.resize(sequence, self.min_frame_size.tolist())
            height, width = sequence.shape[2:]

        def _convolve():
            # convole each sample sequentially to avoid gpu memory issues
            def conv(x):
                return self.conv(x.unsqueeze(1))

            return torch.cat(
                tuple(map(conv, torch.unbind(F.pad(sequence, self.pad), dim=0))), dim=0
            )

        if ftype == "mean":
            out = _convolve() / self.kernel_size**2
        elif ftype == "sum":
            out = _convolve()
        elif ftype == "median":
            out = median(sequence, self.kernel_size)
        else:
            raise ValueError("ftype must be 'sum', 'mean', or 'median." f"Is {ftype}.")

        if hex_sample is True:
            return self.hex_render(out).reshape(samples, frames, 1, -1)

        return out.reshape(samples, frames, height, width)

    def hex_render(self, sequence: torch.Tensor) -> torch.Tensor:
        """Sample receptor locations from a sequence of cartesian frames.

        Args:
            sequence: Cartesian movie sequences of shape (samples, frames, height, width).

        Returns:
            torch.Tensor: Shape (samples, frames, 1, hexals).

        Note:
            Resizes the sequence to the minimum frame size if necessary.
        """
        h, w = sequence.shape[2:]
        if (self.min_frame_size > torch.tensor([h, w])).any():
            sequence = ttf.resize(sequence, self.min_frame_size.tolist())
            h, w = sequence.shape[2:]
        c = self.receptor_centers + torch.tensor([h // 2, w // 2])
        out = sequence[:, :, c[:, 0], c[:, 1]]
        return out.view(*sequence.shape[:2], 1, -1)

    def illustrate(self) -> plt.Figure:
        """Illustrate the receptive field centers and the hexagonal sampling.

        Returns:
            plt.Figure: Matplotlib figure object.
        """
        figsize = [2, 2]
        fontsize = 5
        y_hc, x_hc = np.array(list(self._receptor_centers())).T

        height, width = self.min_frame_size.cpu().numpy()
        x_img, y_img = np.array(
            list(
                product(
                    np.arange(-width / 2, width / 2),
                    np.arange(-height / 2, height / 2),
                )
            )
        ).T

        r = np.sqrt(2) * self.kernel_size / 2

        vertices = []
        angles = [45, 135, 225, 315, 405]
        for _y_c, _x_c in zip(y_hc, x_hc):
            _vertices = []
            for angle in angles:
                offset = r * np.exp(np.radians(angle) * 1j)
                _vertices.append([_y_c + offset.real, _x_c + offset.imag])
            vertices.append(_vertices)
        vertices = np.transpose(vertices, (1, 2, 0))

        fig, ax = init_plot(figsize=figsize, fontsize=fontsize)
        ax.scatter(x_hc, y_hc, color="#00008B", zorder=1, s=0.5)
        ax.scatter(x_img, y_img, color="#34ebd6", s=0.1, zorder=0)

        for h in range(len(x_hc)):
            for i in range(4):
                y1, x1 = vertices[i, :, h]  # x1, y1: (n_hexagons)
                y2, x2 = vertices[i + 1, :, h]
                ax.plot([x1, x2], [y1, y2], c="black", lw=0.25)

        ax.set_xlim(-width / 2, width / 2)
        ax.set_ylim(-height / 2, height / 2)
        rm_spines(ax)
        # fig.tight_layout()
        return fig


# ----- HexEye (slower, more precise) ----------------------------------------


class HexEye:
    """Hexagonal eye model for more precise rendering.

    Args:
        n_ommatidia: Number of ommatidia in the eye. Must currently fill a regular
            hex grid.
        ppo: Pixels per ommatidium.
        monitor_height_px: Monitor height in pixels.
        monitor_width_px: Monitor width in pixels.
        device: Computation device.
        dtype: Data type for computations.

    Attributes:
        monitor_width_px (int): Monitor width in pixels.
        monitor_height_px (int): Monitor height in pixels.
        is_inside (torch.Tensor): Boolean mask for pixels inside hexagons.
        n_ommatidia (int): Number of ommatidia in the eye.
        omm_width_rad (float): Ommatidium width in radians.
        omm_height_rad (float): Ommatidium height in radians.
        ppo (int): Pixels per ommatidium.
        n_hex_circfer (float): Number of hexagons in the circumference.
        device (torch.device): Computation device.
        dtype (torch.dtype): Data type for computations.
    """

    def __init__(
        self,
        n_ommatidia: int = 721,
        ppo: int = 25,
        monitor_height_px: Optional[int] = None,
        monitor_width_px: Optional[int] = None,
        device: torch.device = flyvis.device,
        dtype: torch.dtype = torch.float16,
    ):
        n_hex_circfer = 2 * (-1 / 2 + np.sqrt(1 / 4 - ((1 - n_ommatidia) / 3))) + 1

        if n_hex_circfer % 1 != 0:
            raise ValueError(f"{n_ommatidia} does not fill a regular hex grid.")

        self.monitor_width_px = monitor_width_px or ppo * int(n_hex_circfer)
        self.monitor_height_px = monitor_height_px or ppo * int(n_hex_circfer)

        x_hc, y_hc, (dist_w, dist_h) = hex_center_coordinates(
            n_ommatidia, self.monitor_width_px, self.monitor_height_px
        )

        x_img, y_img = np.array(
            list(
                product(
                    np.arange(self.monitor_width_px),
                    np.arange(self.monitor_height_px),
                )
            )
        ).T

        dist_to_edge = (dist_w + dist_h) / 4

        _, self.is_inside = is_inside_hex(
            torch.tensor(y_img, dtype=dtype, device=device),
            torch.tensor(x_img, dtype=dtype, device=device),
            torch.tensor(x_hc, dtype=dtype, device=device),
            torch.tensor(y_hc, dtype=dtype, device=device),
            torch.tensor(dist_to_edge, dtype=dtype, device=device),
            torch.tensor(np.radians(0), dtype=dtype, device=device),
            device=device,
            dtype=dtype,
        )
        self.kernel_sum = self.is_inside.sum(dim=0)
        # Clean up excessive memory usage.
        if device != "cpu":
            torch.cuda.empty_cache()
        self.n_ommatidia = n_ommatidia
        self.omm_width_rad = np.radians(5.8)
        self.omm_height_rad = np.radians(5.8)
        self.ppo = ppo or int(
            (self.monitor_width_px + self.monitor_width_px) / (2 * n_hex_circfer)
        )
        self.n_hex_circfer = n_hex_circfer
        self.device = device
        self.dtype = dtype

    def __call__(
        self,
        stim: torch.Tensor,
        mode: Literal["mean", "median", "sum"] = "mean",
        n_chunks: int = 1,
    ) -> torch.Tensor:
        """Process stimulus through the hexagonal eye model.

        Args:
            stim: Input stimulus tensor (n_frames, height, width) or
                (n_frames, height * width). Height and width must correspond to the
                monitor size.
            mode: Processing mode.
            n_chunks: Number of chunks to process the stimulus in.

        Returns:
            torch.Tensor: Processed stimulus.
        """
        shape = stim.shape
        if mode not in ["mean", "median", "sum"]:
            raise ValueError

        if len(stim.shape) == 3:
            h, w = stim.shape[1:]
            # resize to monitor size if necessary
            if h < self.monitor_height_px or w < self.monitor_width_px:
                stim = ttf.resize(stim, [self.monitor_height_px, self.monitor_width_px])
            stim = stim.reshape(shape[0], -1)

        try:
            if mode == "median":
                n_pixels = shape[1]
                stim = median(
                    stim.view(-1, self.monitor_height_px, self.monitor_width_px)
                    .float()
                    .to(self.device),
                    int(np.sqrt(self.ppo)),
                ).view(-1, n_pixels)
            elif mode == "sum":
                return (stim[:, :, None] * self.is_inside).sum(dim=1)
            elif mode == "mean":
                return (stim[:, :, None] * self.is_inside).sum(dim=1) / self.kernel_sum
            else:
                raise ValueError(f"Invalid mode: {mode}")

        except RuntimeError as e:
            if "memory" not in str(e):
                raise e
            if "CUDA" in str(e):
                torch.cuda.empty_cache()
            if n_chunks > shape[0]:
                raise ValueError from e

            chunks = torch.chunk(stim, max(n_chunks, 1), dim=0)

            def map_fn(chunk):
                return self(chunk, mode=mode, n_chunks=n_chunks + 1)

            return torch.cat(tuple(map(map_fn, chunks)), dim=0)

    # Rendering functions
    # TODO: move to separate file and make agnostic of the eye model

    def render_bar(
        self,
        bar_width_rad: float,
        bar_height_rad: float,
        bar_loc_theta: float,
        bar_loc_phi: float,
        n_bars: int,
        bar_intensity: float,
        bg_intensity: float,
        moving_angle: float,
        cartesian: bool = False,
        mode: Literal["mean", "median", "sum"] = "mean",
    ) -> Union[np.ndarray, torch.Tensor]:
        """Render a bar stimulus.

        Args:
            bar_width_rad: Width of bars in radians.
            bar_height_rad: Height of bars in radians.
            bar_loc_theta: Horizontal location of bars in radians.
            bar_loc_phi: Vertical location of bars in radians.
            n_bars: Number of bars.
            bar_intensity: Intensity of the bar.
            bg_intensity: Intensity of the background.
            moving_angle: Rotation angle in degrees.
            cartesian: If True, return cartesian coordinates.
            mode: Processing mode.

        Returns:
            Union[np.ndarray, torch.Tensor]: Generated bar stimulus.
        """
        bar_width_px = int(bar_width_rad / self.omm_width_rad * self.ppo)
        bar_height_px = int(bar_height_rad / self.omm_height_rad * self.ppo)
        bar_loc_horizontal_px = int(
            self.monitor_width_px * bar_loc_theta / np.radians(180)
        )
        bar_loc_vertical_px = int(self.monitor_height_px * bar_loc_phi / np.radians(180))

        bar = render_bars_cartesian(
            self.monitor_height_px,
            self.monitor_width_px,
            bar_width_px,
            bar_height_px,
            bar_loc_horizontal_px,
            bar_loc_vertical_px,
            n_bars,
            bar_intensity,
            bg_intensity,
            moving_angle,
        )
        if cartesian:
            return bar
        return self(torch.tensor(bar.flatten(), device=self.device)[None], mode)

    def render_grating(
        self,
        period_rad: float,
        phase_rad: float,
        intensity: float,
        bg_intensity: float,
        moving_angle: float,
        width_rad: Optional[float] = None,
        height_rad: Optional[float] = None,
        cartesian: bool = False,
        mode: Literal["mean", "median", "sum"] = "mean",
    ) -> Union[np.ndarray, torch.Tensor]:
        """Render a grating stimulus.

        Args:
            period_rad: Period of the grating in radians.
            phase_rad: Phase of the grating in radians.
            intensity: Intensity of the grating.
            bg_intensity: Intensity of the background.
            moving_angle: Rotation angle in degrees.
            width_rad: Width of the grating in radians.
            height_rad: Height of the grating in radians.
            cartesian: If True, return cartesian coordinates.
            mode: Processing mode.

        Returns:
            Union[np.ndarray, torch.Tensor]: Generated grating stimulus.
        """
        period_px = int(period_rad / self.omm_width_rad * self.ppo)
        phase_px = int(phase_rad / self.omm_width_rad * self.ppo)

        height_rad_px = None
        if height_rad:
            height_rad_px = int(height_rad / self.omm_height_rad * self.ppo)

        width_rad_px = None
        if width_rad:
            width_rad_px = int(width_rad / self.omm_width_rad * self.ppo)

        grating = render_gratings_cartesian(
            self.monitor_height_px,
            self.monitor_width_px,
            period_px,
            intensity,
            bg_intensity,
            grating_phase_px=phase_px,
            rotate=moving_angle,
            grating_height_px=height_rad_px,
            grating_width_px=width_rad_px,
        )
        if cartesian:
            return grating
        return self(torch.tensor(grating.flatten(), device=self.device)[None], mode)

    def render_grating_offsets(
        self,
        period_rad: float,
        intensity: float,
        bg_intensity: float,
        moving_angle: float,
        width_rad: Optional[float] = None,
        height_rad: Optional[float] = None,
        cartesian: bool = False,
        mode: Literal["mean", "median", "sum"] = "mean",
    ) -> Union[np.ndarray, torch.Tensor]:
        """Render grating stimuli with a range of offsets.

        Args:
            period_rad: Period of the grating in radians.
            intensity: Intensity of the grating.
            bg_intensity: Intensity of the background.
            moving_angle: Rotation angle in degrees.
            width_rad: Width of the grating in radians.
            height_rad: Height of the grating in radians.
            cartesian: If True, return cartesian coordinates.
            mode: Processing mode.

        Returns:
            Union[np.ndarray, torch.Tensor]: Generated grating stimuli with offsets.
        """
        dphase_px = np.radians(
            5.8 / 2
        )  # half ommatidia width - corresponds to led width of 2.25 degree
        n_offsets = np.ceil(period_rad / dphase_px).astype(int)
        gratings = []
        for offset in range(n_offsets):
            gratings.append(
                self.render_grating(
                    period_rad,
                    offset * dphase_px,
                    intensity,
                    bg_intensity,
                    moving_angle,
                    width_rad=width_rad,
                    height_rad=height_rad,
                    cartesian=cartesian,
                    mode=mode,
                )
            )
        if cartesian:
            return np.array(gratings)
        return torch.cat(gratings, dim=0)

    def render_offset_bars(
        self,
        bar_width_rad: float,
        bar_height_rad: float,
        n_bars: int,
        offsets: List[float],
        bar_intensity: float,
        bg_intensity: float,
        moving_angle: float,
        bar_loc_horizontal: float = np.radians(90),
        bar_loc_vertical: float = np.radians(90),
        mode: Literal["mean", "median", "sum"] = "mean",
    ) -> torch.Tensor:
        """Render bars with a range of offsets.

        Args:
            bar_width_rad: Width of bars in radians.
            bar_height_rad: Height of bars in radians.
            n_bars: Number of bars.
            offsets: Offsets of bars wrt. the center in radians.
            bar_intensity: Intensity of the bar.
            bg_intensity: Intensity of the background.
            moving_angle: Rotation angle in degrees.
            bar_loc_horizontal: Horizontal location of bars in radians.
            bar_loc_vertical: Vertical location of bars in radians.
            mode: Processing mode.

        Returns:
            torch.Tensor: Generated offset bars.
        """
        flashes = []
        for offset in offsets:
            flashes.append(
                self.render_bar(
                    bar_width_rad,
                    bar_height_rad,
                    bar_loc_horizontal + offset,
                    bar_loc_vertical,
                    n_bars,
                    bar_intensity,
                    bg_intensity,
                    moving_angle,
                    mode=mode,
                )
            )
        return torch.cat(flashes, dim=0)

    def render_bar_movie(
        self,
        t_stim: float,
        dt: float,
        bar_width_rad: float,
        bar_height_rad: float,
        n_bars: int,
        offsets: List[float],
        bar_intensity: float,
        bg_intensity: float,
        moving_angle: float,
        t_pre: float = 0.0,
        t_between: float = 0.0,
        t_post: float = 0.0,
        bar_loc_horizontal: float = np.radians(90),
        bar_loc_vertical: float = np.radians(90),
    ) -> torch.Tensor:
        """Render moving bars.

        Args:
            t_stim: Stimulus duration.
            dt: Temporal resolution.
            bar_width_rad: Width of bars in radians.
            bar_height_rad: Height of bars in radians.
            n_bars: Number of bars.
            offsets: Offsets of bars wrt. the center in radians.
            bar_intensity: Intensity of the bar.
            bg_intensity: Intensity of the background.
            moving_angle: Rotation angle in degrees.
            t_pre: Grey pre stimulus duration.
            t_between: Grey between offset stimulus duration.
            t_post: Grey post stimulus duration.
            bar_loc_horizontal: Horizontal location of bars in radians.
            bar_loc_vertical: Vertical location of bars in radians.

        Returns:
            torch.Tensor: Generated moving bars.
        """
        pre_frames = round(t_pre / dt)
        stim_frames = round(t_stim / (len(offsets) * dt))
        if stim_frames == 0:
            raise ValueError(
                f"stimulus time {t_stim}s not sufficient to sample {len(offsets)} "
                "offsets at {dt}s"
            )
        between_frames = round(t_between / dt)
        post_frames = round(t_post / dt)

        flashes = []
        if pre_frames:
            flashes.append(torch.ones([pre_frames, self.n_ommatidia]) * bg_intensity)

        for i, offset in enumerate(offsets):
            flash = self.render_bar(
                bar_width_rad,
                bar_height_rad,
                bar_loc_horizontal + offset,
                bar_loc_vertical,
                n_bars,
                bar_intensity,
                bg_intensity,
                moving_angle,
            )
            flashes.append(flash.repeat(stim_frames, 1))

            if between_frames and i < len(offsets) - 1:
                flashes.append(
                    torch.ones([between_frames, self.n_ommatidia]) * bg_intensity
                )
        if post_frames:
            flashes.append(torch.ones([post_frames, self.n_ommatidia]) * bg_intensity)
        return torch.cat(flashes, dim=0)

    def illustrate(
        self, figsize: List[int] = [5, 5], fontsize: int = 5
    ) -> Tuple[plt.Figure, plt.Axes]:
        """Illustrate the hexagonal eye model.

        Args:
            figsize: Figure size.
            fontsize: Font size for the plot.

        Returns:
            Tuple[plt.Figure, plt.Axes]: Matplotlib figure and axes objects.
        """
        x_hc, y_hc, (dist_w, dist_h) = hex_center_coordinates(
            self.n_ommatidia, self.monitor_width_px, self.monitor_height_px
        )

        x_img, y_img = np.array(
            list(
                product(
                    np.arange(self.monitor_width_px),
                    np.arange(
                        self.monitor_height_px,
                    ),
                )
            )
        ).T

        dist_to_edge = (dist_w + dist_h) / 4

        vertices, _ = is_inside_hex(
            torch.tensor(y_img, dtype=self.dtype),
            torch.tensor(x_img, dtype=self.dtype),
            torch.tensor(x_hc, dtype=self.dtype),
            torch.tensor(y_hc, dtype=self.dtype),
            torch.tensor(dist_to_edge, dtype=self.dtype),
            torch.tensor(np.radians(0), dtype=self.dtype),
        )
        vertices = vertices.cpu()
        fig, ax = init_plot(figsize=figsize, fontsize=fontsize)
        ax.scatter(x_hc, y_hc, color="#eb4034", zorder=1)
        ax.scatter(x_img, y_img, color="#34ebd6", s=0.5, zorder=0)

        for h in range(self.n_ommatidia):
            for i in range(6):
                x1, y1 = vertices[i, :, h]  # x1, y1: (n_hexagons)
                x2, y2 = vertices[i + 1, :, h]
                ax.plot([x1, x2], [y1, y2], c="black")

        ax.set_xlim(0, self.monitor_width_px)
        ax.set_ylim(0, self.monitor_height_px)
        rm_spines(ax)
        # fig.tight_layout()
        return fig, ax



"""Modules for decoding the DMN activity."""

import logging
from typing import Dict, List, Optional, Union

import numpy as np
import torch
import torch.nn.functional as nnf
from datamate import Namespace
from torch import nn

from flyvis import device
from flyvis.connectome import ConnectomeFromAvgFilters
from flyvis.utils.activity_utils import LayerActivity
from flyvis.utils.hex_utils import get_hex_coords
from flyvis.utils.nn_utils import n_params

logging = logging.getLogger(__name__)

__all__ = ["ActivityDecoder", "DecoderGAVP", "init_decoder"]


class ActivityDecoder(nn.Module):
    """
    Base class for decoding DMN activity.

    Args:
        connectome: Connectome directory with output_cell_types.

    Attributes:
        dvs_channels (LayerActivity): Dictionary of DVS channels.
        num_parameters (NumberOfParams): Number of parameters in the model.
        u (torch.Tensor): u-coordinates of hexagonal grid.
        v (torch.Tensor): v-coordinates of hexagonal grid.
        H (int): Height of the hexagonal grid.
        W (int): Width of the hexagonal grid.
    """

    dvs_channels: Union[Dict[str, torch.Tensor], LayerActivity]

    def __init__(self, connectome: ConnectomeFromAvgFilters):
        super().__init__()
        self.dvs_channels = LayerActivity(None, connectome, use_central=False)
        self.num_parameters = n_params(self)
        radius = connectome.config.extent
        self.u, self.v = get_hex_coords(radius)
        self.u -= self.u.min()
        self.v -= self.v.min()
        self.H, self.W = self.u.max() + 1, self.v.max() + 1

    def forward(self, activity: torch.Tensor) -> Dict[str, torch.Tensor]:
        """
        Forward pass of the ActivityDecoder.

        Args:
            activity: Tensor of shape (n_samples, n_frames, n_cells).

        Returns:
            Dictionary of tensors with shape
            (n_samples, n_frames, output_cell_types, n_hexals).
        """
        self.dvs_channels.update(activity)
        return self.dvs_channels


class GlobalAvgPool(nn.Module):
    """Returns the average over the last dimension."""

    def forward(self, x):
        return x.mean(dim=-1)


class Conv2dConstWeight(nn.Conv2d):
    """
    PyTorch's Conv2d layer with optional constant weight initialization.

    Args:
        in_channels: Number of input channels.
        out_channels: Number of output channels.
        kernel_size: Size of the convolutional kernel.
        const_weight: Optional constant value for weight initialization.
            If None, the standard PyTorch initialization is used.
        stride: Stride of the convolution.
        padding: Padding added to input.
        **kwargs: Additional keyword arguments for Conv2d.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        const_weight: Optional[float] = None,
        stride: int = 1,
        padding: int = 0,
        **kwargs,
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            stride=stride,
            padding=padding,
            **kwargs,
        )
        if const_weight is not None and self.weight is not None:
            self.weight.data.fill_(const_weight)
        if const_weight is not None and self.bias is not None:
            self.bias.data.fill_(const_weight)


class Conv2dHexSpace(Conv2dConstWeight):
    """
    Convolution with regularly, hexagonally shaped filters (in cartesian map storage).

    Reference to map storage:
    https://www.redblobgames.com/grids/hexagons/#map-storage

    Info:
        kernel_size must be odd!

    Args:
        in_channels: Number of input channels.
        out_channels: Number of output channels.
        kernel_size: Size of the convolutional kernel.
        const_weight: Optional constant value for weight initialization.
            If None, the standard PyTorch initialization is used.
        stride: Stride of the convolution.
        padding: Padding added to input.
        **kwargs: Additional keyword arguments for Conv2d.

    Attributes:
        mask: Mask for hexagonal convolution.
        _filter_to_hex: Whether to apply hexagonal filter.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        const_weight: Optional[float] = 1e-3,
        stride: int = 1,
        padding: int = 0,
        **kwargs,
    ):
        super().__init__(
            in_channels,
            out_channels,
            kernel_size,
            const_weight,
            stride=stride,
            padding=padding,
            **kwargs,
        )

        if not kernel_size % 2:
            raise ValueError(f"{kernel_size} is even. Must be odd.")
        if kernel_size > 1:
            u, v = get_hex_coords(kernel_size // 2)
            u -= u.min()
            v -= v.min()
            mask = np.zeros(tuple(self.weight.shape))
            mask[:, :, u, v] = 1
            self.mask = torch.tensor(mask, device="cpu")
            self.weight.data.mul_(self.mask.to(device))
            self._filter_to_hex = True
        else:
            self._filter_to_hex = False

    def filter_to_hex(self):
        """Apply hexagonal filter to weights."""
        self.weight.data.mul_(self.mask.to(device))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the Conv2dHexSpace layer.

        Args:
            x: Input tensor.

        Returns:
            Output tensor after hexagonal convolution.
        """
        if self._filter_to_hex:
            self.filter_to_hex()
        return super().forward(x)


class DecoderGAVP(ActivityDecoder):
    """
    Fully convolutional decoder with optional global average pooling.

    Args:
        connectome: Connectome directory.
        shape: List of channel sizes for each layer.
        kernel_size: Size of the convolutional kernel.
        p_dropout: Dropout probability.
        batch_norm: Whether to use batch normalization.
        n_out_features: Number of output features.
        const_weight: Constant value for weight initialization.
        normalize_last: Whether to normalize the last layer.
        activation: Activation function to use.

    Attributes:
        _out_channels: Number of output channels before reshaping.
        out_channels: Total number of output channels.
        n_out_features: Number of output features.
        base: Base convolutional layers.
        decoder: Decoder convolutional layers.
        head: Head layers for global average pooling.
        normalize_last: Whether to normalize the last layer.
        num_parameters: Number of parameters in the model.
    """

    def __init__(
        self,
        connectome: ConnectomeFromAvgFilters,
        shape: List[int],
        kernel_size: int,
        p_dropout: float = 0.5,
        batch_norm: bool = True,
        n_out_features: Optional[int] = None,
        const_weight: Optional[float] = None,
        normalize_last: bool = True,
        activation: str = "Softplus",
    ):
        super().__init__(connectome)
        p = int((kernel_size - 1) / 2)
        in_channels = len(connectome.output_cell_types)
        out_channels = shape[-1]
        self._out_channels = out_channels
        self.out_channels = (
            out_channels * n_out_features if n_out_features is not None else out_channels
        )
        self.n_out_features = n_out_features

        self.base = []
        for c in shape[:-1]:
            if c == 0:
                continue
            self.base.append(
                Conv2dHexSpace(
                    in_channels,
                    c,
                    kernel_size,
                    const_weight=const_weight,
                    padding=p,
                )
            )
            if batch_norm:
                self.base.append(nn.BatchNorm2d(c))
            self.base.append(getattr(nn, activation)())
            if p_dropout:
                self.base.append(nn.Dropout(p_dropout))
            in_channels = c
        self.base = nn.Sequential(*self.base)

        self.decoder = []
        if len(self.base) == 0 and batch_norm:
            self.decoder.append(nn.BatchNorm2d(in_channels))
        self.decoder.append(
            Conv2dHexSpace(
                in_channels,
                self.out_channels + 1 if normalize_last else self.out_channels,
                kernel_size,
                const_weight=const_weight,
                padding=p,
            )
        )
        self.decoder = nn.Sequential(*self.decoder)

        self.n_out_features = n_out_features
        self.head = []
        if n_out_features is not None:
            self.head.append(GlobalAvgPool())
        self.head = nn.Sequential(*self.head)

        self.normalize_last = normalize_last

        self.num_parameters = n_params(self)
        logging.info(f"Initialized decoder with {self.num_parameters} parameters.")
        logging.info(repr(self))

    def forward(self, activity: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the DecoderGAVP.

        Args:
            activity: Input activity tensor.

        Returns:
            Decoded output tensor.
        """
        self.dvs_channels.update(activity)
        # Ensure that the outputs of the dvs-model are rectified potentials.
        x = nnf.relu(self.dvs_channels.output)

        # (n_frames, #samples, #outputneurons, n_hexals)
        n_samples, n_frames, in_channels, n_hexals = x.shape

        # Store hexals in square map.
        # (n_frames, #samples, #outputneurons, H, W)
        x_map = torch.zeros([n_samples, n_frames, in_channels, self.H, self.W])
        x_map[:, :, :, self.u, self.v] = x

        # Concatenate actual batch dimension with the frame dimension.
        # torch.flatten(x_map, 0, 1)  # (#samples*n_frames, #outputneurons, H, W)
        x_map = x_map.view(-1, in_channels, self.H, self.W)

        # Run decoder.
        # (n_frames*#samples, out_channels + 1, H, W)
        out = self.decoder(self.base(x_map))

        if self.normalize_last:
            # Do some normalization with the additional channel.
            # (n_frames*#samples, out_channels, H, W)
            out = out[:, : self.out_channels] / (
                nnf.softplus(out[:, self.out_channels :]) + 1
            )

        # Bring back into shape: # (#samples, n_frames, out_channels, n_hexals)
        out = out.view(n_samples, n_frames, self.out_channels, self.H, self.W)[
            :, :, :, self.u, self.v
        ]

        if self.n_out_features is not None:
            out = self.head(out).view(
                n_samples, n_frames, self._out_channels, self.n_out_features
            )

        return out


def init_decoder(
    decoder_config: Namespace, connectome: ConnectomeFromAvgFilters
) -> nn.Module:
    """
    Initialize a decoder based on the provided configuration.

    Args:
        decoder_config: Configuration for the decoder.
        connectome: Connectome directory.

    Returns:
        Initialized decoder module.
    """
    decoder_config = decoder_config.deepcopy()
    _type = decoder_config.pop("type")
    decoder_type = globals()[_type]
    decoder_config.update(dict(connectome=connectome))
    return decoder_type(**decoder_config)
